from __future__ import annotations

import numpy as np
import pandas as pd
from numpy import timedelta64

from retentioneering.eventstream.types import EventstreamType


class _Describe:
    OUT_COLS = ("value",)
    INDEX_NAMES = ("category", "metric")
    TIME_ROUND_UNIT = "s"

    def __init__(
        self, eventstream: EventstreamType, session_col: str = "session_id", raw_events_only: bool = False
    ) -> None:
        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp
        self.type_col = self.__eventstream.schema.event_type
        self.session_col = session_col

        self.overall_stats = [
            ["overall"],
            ["unique_users", "unique_events", "eventstream_start", "eventstream_end", "eventstream_length"],
        ]
        self.time_events_stats = [["path_length_time", "path_length_steps"], ["mean", "std", "median", "min", "max"]]

        self.df = self.__eventstream.to_dataframe(copy=True)
        self.has_session_col: bool = False
        if self.session_col in self.df.columns:
            self.has_session_col = True

        if raw_events_only:
            self.df = self.df[self.df[self.type_col].isin(["raw"])]

    def _calc_statistics(self, agg_col: str) -> list[np.timedelta64 | int | float]:
        df_agg = self.df.groupby(agg_col).agg({self.time_col: ["min", "max"], self.event_col: ["count"]}).reset_index()
        time_diff_user = df_agg[(self.time_col, "max")] - df_agg[(self.time_col, "min")]
        mean_time_agg_col = time_diff_user.mean().round(self.TIME_ROUND_UNIT)  # type: ignore
        median_time_agg_col = time_diff_user.median().round(self.TIME_ROUND_UNIT)  # type: ignore
        std_time_agg_col = time_diff_user.std().round(self.TIME_ROUND_UNIT)  # type: ignore
        min_length_time_agg_col = time_diff_user.min().round(self.TIME_ROUND_UNIT)  # type: ignore
        max_length_time_agg_col = time_diff_user.max().round(self.TIME_ROUND_UNIT)  # type: ignore

        event_count_agg_col = df_agg[(self.event_col, "count")]
        mean_event_agg_col = round(event_count_agg_col.mean(), 2)  # type: ignore
        median_event_agg_col = event_count_agg_col.median()
        std_event_agg_col = round(event_count_agg_col.std(), 2)  # type: ignore
        min_event_length_agg_col = event_count_agg_col.min()
        max_event_length_agg_col = event_count_agg_col.max()

        values_time_events = [
            mean_time_agg_col,
            std_time_agg_col,
            median_time_agg_col,
            min_length_time_agg_col,
            max_length_time_agg_col,
            mean_event_agg_col,
            std_event_agg_col,
            median_event_agg_col,
            min_event_length_agg_col,
            max_event_length_agg_col,
        ]
        return values_time_events  # type: ignore

    def _output_df_construction(
        self, values_overall: list[timedelta64 | int | float], values_time_events: list[timedelta64 | int | float]
    ) -> pd.DataFrame:
        overall_index = pd.MultiIndex.from_product(self.overall_stats, names=self.INDEX_NAMES)
        time_events_index = pd.MultiIndex.from_product(self.time_events_stats, names=self.INDEX_NAMES)

        # TODO внезапно в старом коде появилась ошибка типов, разобраться
        df_overall = pd.DataFrame(data=values_overall, index=overall_index, columns=self.OUT_COLS)  # type: ignore
        df_time_events = pd.DataFrame(data=values_time_events, index=time_events_index, columns=self.OUT_COLS)  # type: ignore

        return pd.concat([df_overall, df_time_events])

    def _values(self) -> pd.DataFrame:
        max_time = self.df[self.time_col].max()
        min_time = self.df[self.time_col].min()

        values_overall = [
            self.df[self.user_col].nunique(),
            self.df[self.event_col].nunique(),
            min_time.round(self.TIME_ROUND_UNIT),
            max_time.round(self.TIME_ROUND_UNIT),
            (max_time - min_time).round(self.TIME_ROUND_UNIT),
        ]

        values_time_events = self._calc_statistics(self.user_col)

        if self.has_session_col:
            self.time_events_stats[0] += ["session_length_time", "session_length_steps"]
            self.overall_stats[1].insert(2, "unique_sessions")

            values_overall.insert(2, self.df[self.session_col].nunique())
            values_time_events += self._calc_statistics(self.session_col)

        return self._output_df_construction(values_overall, values_time_events)  # type: ignore


__all__ = ("_Describe",)



from __future__ import annotations

from typing import Any, Literal, Optional, Tuple

import pandas as pd
from typing_extensions import TypeGuard

from retentioneering.eventstream.types import (
    EventstreamSchemaType,
    EventstreamType,
    SegmentFilterExpr,
    SplitExpr,
    UserGroupsNamesType,
    UserGroupsType,
)

SEGMENT_TYPE = "segment"
SEGMENT_DELIMITER = "::"
SEGM_NAME_COL = "segment_name"
SEGM_VALUE_COL = "segment_value"

OUTER_LITERAL = "_OUTER_"
ALL_LITERAL = "_ALL_"


def _create_segment_event(event_col: pd.Series, name: str) -> pd.Series:
    return name + SEGMENT_DELIMITER + event_col.astype(str)


def _get_segment_mask(
    df: pd.DataFrame, schema: EventstreamSchemaType, name: Optional[str] = None, check_existence: bool = True
) -> pd.Series[bool]:
    if not name:  # return all segments
        mask = df[schema.event_type] == SEGMENT_TYPE
    else:
        mask = (df[schema.event_type] == SEGMENT_TYPE) & (_extract_segment_keys(df.loc[:, schema.event_name]) == name)

    if check_existence and (~mask).all():  # type: ignore
        raise KeyError(f"segment with name {name} is non-existent!")

    return mask


def _extract_segment_values(col: pd.Series) -> pd.Series:
    return col.astype(str).str.split(SEGMENT_DELIMITER).str[-1]


def _extract_segment_keys(col: pd.Series) -> pd.Series:
    return col.astype(str).str.split(SEGMENT_DELIMITER).str[0]


def _get_segments_last_values(df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
    event_type_col = schema.event_type
    user_col = schema.user_id
    event_col = schema.event_name

    new_df = df[df[event_type_col] == SEGMENT_TYPE][[user_col, event_col]].copy()
    new_df["segment_name"] = _extract_segment_keys(new_df.loc[:, event_col])

    return new_df.groupby([user_col, "segment_name"]).tail(1)


def _calculate_segment_col(df: pd.DataFrame, schema: EventstreamSchemaType, name: str) -> pd.Series[Any]:
    segm_mask = _get_segment_mask(df, schema, name)  # todo @dakhaytin - refactor this check
    tmp = df[[schema.user_id, schema.event_index]].copy()

    tmp["is_current_segment"] = _extract_segment_keys(df.loc[:, schema.event_name]) == name
    tmp["segment_cumsum"] = tmp.groupby(schema.user_id)["is_current_segment"].cumsum()
    tmp["segment_cumsum"] = tmp[schema.user_id].astype(str) + tmp["event_index"].map(
        tmp.groupby("event_index")["segment_cumsum"].max()
    ).astype(
        str
    )  # fill gaps for segment events before target segment

    mask = tmp["is_current_segment"]
    tmp.loc[mask, "segment_value"] = _extract_segment_values(df.loc[mask][schema.event_name])
    tmp[name] = tmp["segment_cumsum"].map(tmp.groupby("segment_cumsum")["segment_value"].first())

    return tmp[name]


def _filter_segment_events(df: pd.DataFrame, schema: EventstreamSchemaType, segment: SegmentFilterExpr) -> None:
    name, values = segment
    if isinstance(values, str):
        values = [values]

    segment_col = _calculate_segment_col(df, schema, name)

    mask = segment_col.isin(values)
    df.drop(df.loc[~mask].index, inplace=True)


def _extract_groups(path_df: pd.Series, value_l: str, value_r: str) -> UserGroupsType:
    if value_l in [OUTER_LITERAL, ALL_LITERAL]:
        raise KeyError("Left value have to be exact value of chosen segment!")

    group_l = path_df.loc[path_df.index == value_l].unique().tolist()

    if value_r == OUTER_LITERAL:
        group_r = path_df.loc[path_df.index != value_l].unique().tolist()
    elif value_r == ALL_LITERAL:
        group_r = path_df.loc[:].unique().tolist()
    else:
        group_r = path_df.loc[path_df.index == value_r].unique().tolist()

    return group_l, group_r  # type: ignore


def _split_segment(
    stream: EventstreamType, split_expr: SplitExpr | UserGroupsType | str
) -> Tuple[UserGroupsType, UserGroupsNamesType]:
    df: pd.Series

    if _is_user_groups(split_expr):
        split_expr = [list(t) for t in split_expr]
        return split_expr, []

    elif isinstance(split_expr, str):  # if binary segment
        df = get_segment_path_map(stream, split_expr, "segment_value")  # type: ignore
        if df.index.nunique() != 2:
            raise ValueError(f"{split_expr} is not a binary segment!")

        value_l, value_r = df.index.unique().tolist()

    elif _is_split_expr(split_expr):  # if split expression
        name, value_l, value_r = split_expr
        df = get_segment_path_map(stream, name, "segment_value")  # type: ignore

    else:
        raise ValueError("Wrong user groups input!")

    groups = _extract_groups(df, value_l, value_r)

    if len(groups[0]) == 0:
        raise ValueError(f"Segment with value {value_l} is empty!")
    elif len(groups[1]) == 0:
        raise ValueError(f"Segment with value {value_r} is empty!")

    return groups, [value_l, value_r]  # type: ignore


def _is_user_groups(split_exp: SplitExpr | UserGroupsType | str) -> TypeGuard[UserGroupsType]:
    check = any(
        [
            all(isinstance(expr, list) for expr in split_exp) and len(split_exp) == 2,
            all(isinstance(expr, set) for expr in split_exp) and len(split_exp) == 2,
        ]
    )

    return check


def _is_split_expr(split_exp: SplitExpr | UserGroupsType | str) -> TypeGuard[SplitExpr]:
    check = (
        len(split_exp) == 3
        and isinstance(split_exp[0], str)
        and all(isinstance(expr, (str, int, float)) for expr in split_exp[1:])
    )
    return check


# PUBLIC API STARTS HERE #
def get_segment_path_map(
    stream: EventstreamType,
    name: str | None = None,
    index: Literal["path_id", "segment_value"] = "path_id",
    resolve_collision: Optional[Literal["majority", "last"]] = None,
) -> pd.DataFrame | pd.Series:
    df = stream.to_dataframe(drop_segment_events=False)
    schema = stream.schema
    user_col = stream.schema.user_id
    event_col = stream.schema.event_name

    df = df.loc[_get_segment_mask(df, schema, name)][[schema.user_id, schema.event_name]].drop_duplicates().copy()
    df[SEGM_NAME_COL] = _extract_segment_keys(df.loc[:, schema.event_name])
    df[schema.event_name] = _extract_segment_values(df.loc[:, schema.event_name])

    if not resolve_collision:
        ...

    elif resolve_collision == "majority":
        df = df.groupby([user_col, SEGM_NAME_COL])[event_col].agg(lambda x: x.value_counts().index[0]).reset_index()

    elif resolve_collision == "last":
        df = df.groupby([user_col, SEGM_NAME_COL])[event_col].last().reset_index()

    else:
        raise KeyError('resolve_collision option should be either "majority" or "last"!')

    df = df.rename(columns={event_col: SEGM_VALUE_COL})

    if not name:
        df = df.sort_values(user_col).reset_index(drop=True)
        df = df[[user_col, SEGM_NAME_COL, SEGM_VALUE_COL]]
        return df

    if index == "path_id":
        df.set_index(user_col, inplace=True)
        return_col = SEGM_VALUE_COL
    elif index == "segment_value":
        df.set_index(SEGM_VALUE_COL, inplace=True)
        return_col = user_col
    else:
        raise KeyError("index must be either path_id or segment_value")

    df.sort_index(inplace=True)
    return df[return_col]


def get_all_segments(stream: EventstreamType, resolve_collision: Literal["majority", "last"]) -> pd.DataFrame:
    user_col = stream.schema.user_id
    event_col = stream.schema.event_name

    all_segments = get_segment_path_map(stream, None, "path_id", resolve_collision)

    all_segments = all_segments.pivot(index=user_col, columns=SEGM_NAME_COL, values=SEGM_VALUE_COL)

    return all_segments



from __future__ import annotations

from dataclasses import dataclass, field

from .edges import EdgeAggregation
from .nodes import NodeAggregation, NodeId
from .threshold import ThresholdValueMap


@dataclass
class StateChanges:
    disabledNodes: list[NodeId] = field(default_factory=list)
    nodesThreshold: ThresholdValueMap = field(default_factory=dict)
    edgesThreshold: ThresholdValueMap = field(default_factory=dict)
    nodesAggregation: list[NodeAggregation] = field(default_factory=list)
    edgesAggregation: list[EdgeAggregation] = field(default_factory=list)



# * Copyright (C) 2020 Maxim Godzi, Retentioneering Team
# * This Source Code Form is subject to the terms of the Retentioneering Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/retentioneering/retentioneering-tools/blob/master/LICENSE.md

from .eventstream import Eventstream
from .schema import EventstreamSchema, RawDataSchema



from __future__ import annotations

import uuid
import warnings
from typing import Any, Callable, Literal, Optional

import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import EnumWidget


def _numeric_values_processing(x: pd.Series) -> int | float:
    return x.mean()


def _string_values_processing(x: pd.Series) -> str | None:
    # check if all the values in the collapsing group are equal
    # NaN values are ignored
    if x.nunique() == 1:
        return x.dropna().max()
    else:
        return None


class CollapseLoopsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.CollapseLoops` class.
    """

    suffix: Optional[Literal["loop", "count"]]
    time_agg: Literal["max", "min", "mean"] = "min"
    _widgets = {
        "time_agg": EnumWidget(optional=False, params=["max", "min", "mean"], default="min"),
    }


@docstrings.get_sections(base="CollapseLoops")  # type: ignore
class CollapseLoops(DataProcessor):
    """
    Find ``loops`` and create new synthetic events in the paths of all users having such sequences.

    A ``loop`` - is a sequence of repetitive events.
    For example *"event1 -> event1"*

    Parameters
    ----------
    suffix : {"loop", "count"}, optional

        - If ``None``, event_name will be event_name without any changes.\n
        For example *"event1 - event1 - event1"* --> event1.

        - If ``loop``, event_name will be event_name_loop.\n
        For example *"event1 - event1 - event1"* --> event1_loop.

        - If ``count``, event_name will be event_name_loop_{number of events}.\n
        For example *"event1 - event1 - event1"* --> event1_loop_3.
    time_agg : {"max", "min", "mean"}, default "min"
        Aggregation method to calculate timestamp values for new groups.

    Returns
    -------
    Eventstream
        Eventstream with:

        - raw events: the returned events will be marked ``_deleted=True`` and soft-deleted from input eventstream.
        - new synthetic events: the returned events will be added to the input eventstream with columns below.

        +------------------------+----------------+--------------------------------------------+
        | **event_name**         | **event_type** | **timestamp**                              |
        +------------------------+----------------+--------------------------------------------+
        | event_name_loop        | group_alias    | min/max/mean(group of repetitive events))  |
        +------------------------+----------------+--------------------------------------------+
        | event_name_loop_{count}| group_alias    | (min/max/mean(group of repetitive events)) |
        +------------------------+----------------+--------------------------------------------+
        | event_name             | group_alias    | (min/max/mean(group of repetitive events)) |
        +------------------------+----------------+--------------------------------------------+


    See Also
    --------
    .StepMatrix : This class provides methods for step matrix calculation and visualization.
    .RawDataSchema : Define schema for ``raw_data`` columns names.

    Notes
    -----
    If an eventstream contains custom columns they will be aggregated in the following way:

    - for numeric columns the mean value will be calculated for each collapsed group. ``None`` values are ignored.
      Supported numeric types are: ``bool``, ``int``, ``float``.
    - for string columns, if all the values to be aggregated in the
      collapsing group are equal, this single value will be returned, otherwise - ``None``.
      ``None`` values in the input data will be ignored.

    See :doc:`Data processors user guide</user_guides/dataprocessors>` and :ref:`Eventstream custom columns'
    explanation<eventstream_raw_data_schema>` for the details.
    """

    params: CollapseLoopsParams
    NUMERIC_DTYPES = ["integer", "floating", "boolean", "mixed-integer-float"]

    @time_performance(
        scope="collapse_loops",
        event_name="init",
    )
    def __init__(self, params: CollapseLoopsParams):
        super().__init__(params=params)

    @time_performance(
        scope="collapse_loops",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        source = df.copy()

        user_col = schema.user_id
        time_col = schema.event_timestamp
        type_col = schema.event_type
        event_col = schema.event_name
        event_index_col = schema.event_index
        custom_cols: list = schema.custom_cols
        event_id = schema.event_id

        suffix = self.params.suffix
        time_agg = self.params.time_agg
        full_agg: dict[str, Any] = {}

        if len(custom_cols) > 0:
            df_custom_cols = df[custom_cols]
            default_agg: dict[str, Callable] = {}

            for col in df_custom_cols.columns:
                if pd.api.types.infer_dtype(df_custom_cols[col]) in self.NUMERIC_DTYPES:
                    default_agg[col] = _numeric_values_processing  # type: ignore
                elif pd.api.types.infer_dtype(df_custom_cols[col]) == "string":
                    default_agg[col] = _string_values_processing  # type: ignore
                else:
                    doc_link = "https://pandas.pydata.org/docs/reference/api/pandas.api.types.infer_dtype.html"
                    message = (
                        f"Column '{col}' with "
                        f"'{pd.api.types.infer_dtype(df_custom_cols[col])}'"
                        f" data type is not supported for collapsing. See {doc_link}"
                    )

                    raise TypeError(message)

            full_agg = default_agg
        else:
            full_agg = {}

        df["grp"] = df[event_col] != df.groupby(user_col)[event_col].shift(1)
        df["cumgroup"] = df.groupby(user_col)["grp"].cumsum()
        df["count"] = df.groupby([user_col, "cumgroup"]).cumcount() + 1
        df["collapsed"] = df.groupby([user_col, "cumgroup", event_col])["count"].transform(max)
        df["collapsed"] = df["collapsed"].apply(lambda x: False if x == 1 else True)

        loops = (
            df[df["collapsed"] == 1]
            .groupby([user_col, "cumgroup", event_col])
            .agg({time_col: time_agg, "count": "max", **full_agg, event_index_col: time_agg})
            .reset_index()
        )
        loops[event_index_col] = loops[event_index_col].astype("int64")
        if suffix == "loop":
            loops[event_col] = loops[event_col].map(str) + "_loop"
        elif suffix == "count":
            loops[event_col] = loops[event_col].map(str) + "_loop_" + loops["count"].map(str)
        loops[type_col] = "group_alias"

        df_to_del = df[df["collapsed"] == 1]

        if len(custom_cols) > 0:
            cols_to_show = [user_col, time_col, type_col, event_col] + custom_cols
            link_url = (
                "https://doc.retentioneering.com/stable/doc/api/preprocessing/data_processors/collapse_loops.html"
            )
            message_template = """
                                    \nThere are NaN values in the {where_nans} columns!
                                    \nThe total amount of NaN values in each column:
                                    \n{total_nan_amount}
                                    \nAs a reference, here are some rows where NaNs occurred:
                                    \n{nan_example}{input_data_add_info}
                                    \nFor more information, see collapse_loops documentation {link}\n
                                """

            if df_to_del[custom_cols].isna().values.sum() > 0:
                cols_with_na = df_to_del[custom_cols].isna().sum()
                rows_to_show = df_to_del[df_to_del[custom_cols].isnull().any(axis=1)].head(3)

                warning_message = message_template.format(
                    where_nans="input",
                    total_nan_amount=cols_with_na,
                    nan_example=rows_to_show[cols_to_show],
                    input_data_add_info="\n\nThese NaN values will be ignored in the further calculation",
                    link=link_url,
                )

                warnings.warn(warning_message)

            if loops[custom_cols].isna().values.sum() > 0:
                cols_with_na = loops[custom_cols].isna().sum()
                rows_to_show = loops[loops[custom_cols].isnull().any(axis=1)].head(3)

                warning_message = message_template.format(
                    where_nans="custom",
                    total_nan_amount=cols_with_na,
                    nan_example=rows_to_show[cols_to_show],
                    input_data_add_info="",
                    link=link_url,
                )

                warnings.warn(warning_message)

        updated = source
        if not df_to_del.empty:
            updated = source[~source[event_id].isin(df_to_del[event_id])]

        loops[schema.event_id] = [uuid.uuid4() for x in range(len(loops))]

        result = pd.concat([updated, loops])
        result = result.drop(["cumgroup", "count"], axis=1, errors="ignore")

        collect_data_performance(
            scope="collapse_loops",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            not_hash_values=["suffix", "time_agg"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



import os

from retentioneering import RETE_CONFIG

from .connector.csv_connector import CSVConnector
from .connector.rete_connector import TrackerMainConnector
from .tracker import Tracker

if csv_name := os.getenv("RETE_TRACKER_CSV_NAME", None):
    tracker = Tracker(connector=CSVConnector(csv_name=csv_name))
else:
    tracker = Tracker(connector=TrackerMainConnector(), enabled=RETE_CONFIG.tracking.is_tracking_allowed)

track = tracker.track
time_performance = tracker.time_performance
collect_data_performance = tracker.collect_data_performance

__all__ = [
    "track",
    "tracker",
    "time_performance",
    "collect_data_performance",
]



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Settings:
    show_weights: bool | None = None
    show_percents: bool | None = None
    show_nodes_names: bool | None = None
    show_all_edges_for_targets: bool | None = None
    show_nodes_without_links: bool | None = None
    show_edge_info_on_hover: bool | None = None
    open_sidebar_by_default: bool | None = None


@dataclass
class SerializedSettings:
    showEdgesWeightsOnCanvas: bool | None = None
    convertWeightsToPercents: bool | None = None
    showNodesNamesOnCanvas: bool | None = None
    doNotFilterTargetNodes: bool | None = None
    showNodesWithoutEdges: bool | None = None
    showEdgesInfoOnHover: bool | None = None
    openSidebarByDefault: bool | None = None



from __future__ import annotations

import json
from typing import Hashable, Tuple

import numpy as np
import pandas as pd

EVENTSTREAM_KEYS = ["shape", "custom_cols", "unique_users", "unique_events", "hash", "eventstream_hist", "index"]


def to_json(data: str) -> dict | float:
    try:
        return json.loads(data)
    except:
        return np.nan


def get_eventstream_before(data: pd.Series) -> dict:
    arg_dict = data["params"]
    eventstream_args: dict = data["eventstream_args"]
    if all([pd.isna(t) for t in eventstream_args.values()]):
        eventstream_args = {}

    if len(arg_dict) == 0:
        return eventstream_args

    data = arg_dict[0].get("performance_info", {})
    data_processor_output: dict

    data_processor_output = data.get("parent", {})  # type:ignore
    default_output = {}
    if all(key in data.keys() for key in EVENTSTREAM_KEYS):
        default_output = {k: v for k, v in data.items() if k in EVENTSTREAM_KEYS}

    result = {**eventstream_args, **(data_processor_output or default_output)}
    return result


def get_eventstream_after(arg_dict: list[dict]) -> dict:
    if len(arg_dict) == 0:
        return {}

    data_processor_output = arg_dict[0].get("performance_info", {}).get("child", {})
    default_output = {}
    if len(arg_dict) > 1 and all(key in arg_dict[1].keys() for key in EVENTSTREAM_KEYS):
        default_output = {k: v for k, v in arg_dict[1].items() if k in EVENTSTREAM_KEYS}

    return data_processor_output or default_output


def get_out_attrs(arg_dict: list[dict]) -> dict:
    first, last = {}, {}
    if len(arg_dict) > 0:
        first = arg_dict[0].get("performance_info", {})
        first.pop("child", None)
        first.pop("parent", None)

        if all(key in first.keys() for key in EVENTSTREAM_KEYS):
            first = {k: v for k, v in first.items() if k not in EVENTSTREAM_KEYS}

    if len(arg_dict) > 1:
        last = arg_dict[1].get("performance_info", {})
        last.pop("child", None)
        last.pop("parent", None)

        if all(key in last.keys() for key in EVENTSTREAM_KEYS):
            last = {k: v for k, v in first.items() if k not in EVENTSTREAM_KEYS}

    return last or first


def set_parent(df: pd.DataFrame, index_col: str = "index", parent_col: str = "parent_index") -> pd.Series:
    """Parses the data as a sequence of brackets and sets the index
    of the parent bracket for all internal data. The algorithm works
    backwards to process the broken sequences.
    """
    df = df.sort_values(by=index_col)
    position = df.shape[0] - 1

    parent_stack, event_stack, parents = [0], [], []
    while position >= 0:
        if parent_stack[-1] == 0 and not (
            df.iloc[position]["event_custom_name"].endswith("_end")
            or df.iloc[position]["event_custom_name"].endswith("_tracker")
        ):
            parents.append(-1)
        else:
            parents.append(parent_stack[-1])
        full_name = f"{df.iloc[position]['scope']}_{df.iloc[position]['event_name']}"
        if df.iloc[position]["event_custom_name"].endswith("_end"):
            event_stack.append(full_name)
            parent_stack.append(df.iloc[position][index_col])
        if df.iloc[position]["event_custom_name"].endswith("_start"):
            while full_name in event_stack:
                parent_stack.pop()
                if event_stack.pop() == full_name:
                    break

        position -= 1

    return pd.Series(parents[::-1], index=df.index, name=parent_col)


def prepare_data(
    df: pd.DataFrame,
    index_col: str = "index",
    parent_col: str = "parent_index",
    full_index: str = "full_index",
    full_parent_index: str = "full_parent_index",
    event_full_name: str = "event_full_name",
    buffer_index_name: str = "_default_index",
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Collapses each bracket group into a single event and
    writes the latest metadata for the current group. The broken sequences are
    excluded from consideration. Returns two dataframes: processed and broken data.
    """
    if df.empty:
        raise ValueError("The data is empty")

    df.sort_values(index_col, inplace=True)
    df[parent_col] = pd.concat([set_parent(group) for name, group in df.groupby(["user_id", "jupyter_kernel_id"])])
    broken = df[df[parent_col] == -1]
    df = df[df.parent_index != -1].copy()
    df[full_index] = df[index_col].apply(str) + df["user_id"] + df["jupyter_kernel_id"]
    df[full_parent_index] = df[parent_col].apply(str) + df["user_id"] + df["jupyter_kernel_id"]
    df[event_full_name] = df["scope"] + "_" + df["event_name"]

    meta_cols = ["eventstream_index", "parent_eventstream_index", "child_eventstream_index", "params"]

    meta = (
        df[(df["event_name"] == "metadata")]
        .groupby(full_parent_index)
        .agg({col: list if col == "params" else "last" for col in meta_cols})
    )
    df.reset_index(inplace=True, names=[buffer_index_name])
    df.set_index(full_index, inplace=True)

    df["params"] = [[] for _ in range(df.shape[0])]
    df.loc[meta.index, meta_cols] = meta[meta_cols]
    df.reset_index(inplace=True)
    df.set_index(buffer_index_name, inplace=True)
    df.index.name = None

    df["session_ix"] = pd.concat([set_sessions(group) for name, group in df.groupby(["user_id"])])

    return df.loc[(df.event_name != "metadata") & ~df.event_custom_name.str.endswith("_start")], broken


def uncover_params(df: pd.DataFrame) -> pd.DataFrame:
    df["args"] = df.params.apply(lambda l: l[0].get("args", {}) if len(l) > 0 else {})
    df["eventstream_args"] = df[["child_eventstream_index", "eventstream_index", "parent_eventstream_index"]].to_dict(
        "records"
    )

    df["eventstream_input"] = df[["eventstream_args", "params"]].apply(get_eventstream_before, axis=1)
    df["eventstream_output"] = df.params.apply(get_eventstream_after)
    df["output_attributes"] = df.params.apply(get_out_attrs)

    return df


def set_sessions(df: pd.DataFrame) -> pd.Series:
    df["is_root_call"] = df["parent_index"] == 0
    df["session_ix"] = df.loc[::-1, "is_root_call"].cumsum()[::-1]
    return df.loc[:, "session_ix"]


def get_inner_calls(
    data: pd.DataFrame, parent_index: Hashable, index_col: str = "full_index", parent_col: str = "full_parent_index"
) -> pd.DataFrame:
    parent_ids = [parent_index]
    inner_ids = []
    while parent_ids:
        current_parent = parent_ids.pop()
        inner_calls = data[data[parent_col] == current_parent]
        parent_ids.extend(inner_calls[index_col])
        inner_ids.append(current_parent)
    return data[data[index_col].isin(inner_ids)]


def aggregate_to_top_calls(df: pd.DataFrame, front_df: pd.DataFrame) -> pd.DataFrame:
    result = df[(df["parent_index"] == 0)].copy()
    cols_to_calc = ["args", "eventstream_input", "eventstream_output", "output_attributes"]

    result.drop(columns=cols_to_calc, inplace=True)
    for col in cols_to_calc:
        grouped = df[df[col] != {}].groupby(["user_id", "session_ix"])[col].last()
        result = result.join(grouped, on=["user_id", "session_ix"])

    timestamp_df = (
        df.groupby(["user_id", "session_ix"])["event_timestamp_ms"]
        .agg(["first", "last"])
        .rename(columns={"first": "timestamp_start", "last": "timestamp_end"})
    )
    result = result.join(timestamp_df, on=["user_id", "session_ix"])

    front_df["timestamp_start"] = front_df["event_timestamp_ms"]
    front_df["timestamp_end"] = front_df["event_timestamp_ms"]
    result = pd.concat([result, front_df]).sort_values("index")

    return result


def process_data(data: pd.DataFrame, only_calls: bool = True) -> pd.DataFrame:
    """Processes raw data from the database."""
    df = data.copy()
    back_df, front_df = separate_data_sources(df)
    back_df["params"] = back_df["params"].apply(to_json)
    back_df = prepare_data(back_df)[0]
    back_df = uncover_params(back_df)

    if only_calls:
        df = aggregate_to_top_calls(back_df, front_df)

    else:
        df = pd.concat([back_df, front_df]).sort_values("index")

    return df


def separate_data_sources(df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
    mask = df["browser"] == "Python Requests"
    back_df = df.loc[mask].copy()
    front_df = df.loc[~mask].copy()

    return back_df, front_df



from __future__ import annotations

from typing import Callable, List, Optional

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class AddNegativeEventsHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(
        scope="add_negative_events",
        event_name="helper",
        event_value="combine",
    )
    def add_negative_events(
        self: EventstreamType, targets: List[str], func: Optional[Callable] = None
    ) -> EventstreamType:
        """
        A method of ``Eventstream`` class that creates new synthetic
        events in paths of all users having the specified events - ``negative_target_RAW_EVENT_NAME``.

        Parameters
        ----------
            %(AddNegativeEvents.parameters)s

        Returns
        -------
        Eventstream
            Input ``eventstream`` with new synthetic events.



        """

        calling_params = {
            "targets": targets,
            "func": func,
        }

        # avoid circular import
        from retentioneering.data_processors_lib import (
            AddNegativeEvents,
            AddNegativeEventsParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        params: dict[str, list[str] | Callable] = {"targets": targets}
        if func:
            params["func"] = func

        node = EventsNode(processor=AddNegativeEvents(params=AddNegativeEventsParams(**params)))  # type: ignore
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="add_negative_events",
            event_name="metadata",
            called_params=calling_params,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )
        return result



class Counter:
    def __init__(self) -> None:
        self.__event_index = 0
        self.__eventstream_index = 0

    def get_event_index(self) -> int:
        self.__event_index += 1
        return self.__event_index

    def get_eventstream_index(self) -> int:
        self.__eventstream_index += 1
        return self.__eventstream_index

    def reload(self) -> None:
        self.__event_index = 0
        self.__eventstream_index = 0


counter = Counter()



from __future__ import annotations

import pandas as pd

from retentioneering.eventstream.types import EventstreamType
from retentioneering.tooling.transition_graph.types import (
    NormType,
    ThresholdValue,
    ThresholdValueMap,
)

IS_OUT_OF_THRESHOLD_COL = "rete_is_out_of_threshold"


class Edgelist:
    edgelist_df: pd.DataFrame
    eventstream: EventstreamType

    _weight_col: str

    def __init__(
        self,
        eventstream: EventstreamType,
        weight_cols: list[str] | None = None,
        edgelist_df: pd.DataFrame | None = None,
    ) -> None:
        self.eventstream = eventstream
        if edgelist_df is not None:
            self.edgelist_df = edgelist_df

        if weight_cols:
            self.weight_cols = weight_cols

    @property
    def weight_col(self) -> str:
        return self._weight_col

    @weight_col.setter
    def weight_col(self, value: str) -> None:
        if not value:
            raise ValueError("Weight col cannot be empty")
        self._weight_col = value

    @property
    def group_col(self) -> str:
        if self.weight_col in (self.eventstream.schema.event_id, self.eventstream.schema.user_id):
            return self.eventstream.schema.user_id
        else:
            return self.weight_col

    @property
    def next_event_col(self) -> str:
        return f"next_{self.eventstream.schema.event_name}"

    def calculate_edgelist(
        self,
        weight_cols: list[str],
        norm_type: NormType = None,
        edges_threshold: ThresholdValueMap | None = None,
    ) -> pd.DataFrame:
        self.weight_cols = weight_cols

        if norm_type not in (None, "full", "node"):
            raise ValueError(f"unknown normalization type: {norm_type}")

        edge_from, edge_to = self.eventstream.schema.event_name, self.next_event_col
        df = self.eventstream.to_dataframe()
        calculated_edgelist: pd.DataFrame = pd.DataFrame()
        for weight_col in weight_cols:
            self.weight_col = weight_col
            edgelist = self._calculate_edgelist_for_selected_weight(
                df=df, norm_type=norm_type, edge_from=edge_from, edge_to=edge_to
            )
            if calculated_edgelist.empty:
                calculated_edgelist = edgelist
            else:
                calculated_edgelist = self._merge_edgelist(calculated_edgelist, edge_from, edge_to, edgelist)

        self.edgelist_df = calculated_edgelist
        self.edgelist_df[IS_OUT_OF_THRESHOLD_COL] = False
        self.update_threshold(edges_threshold=edges_threshold)

        return calculated_edgelist

    def _merge_edgelist(
        self, calculated_edgelist: pd.DataFrame, edge_from: str, edge_to: str, edgelist: pd.DataFrame
    ) -> pd.DataFrame:
        calculated_edgelist = pd.merge(
            calculated_edgelist,
            edgelist,
            how="outer",
            left_on=[edge_from, edge_to],
            right_on=[edge_from, edge_to],
            suffixes=["", "_del"],
        )
        calculated_edgelist = calculated_edgelist[
            [c for c in calculated_edgelist.columns if not c.endswith("_del")]  # type: ignore
        ]
        return calculated_edgelist

    def _calculate_edgelist_for_selected_weight(
        self, df: pd.DataFrame, norm_type: NormType, edge_from: str, edge_to: str
    ) -> pd.DataFrame:
        possible_transitions = (
            df.assign(**{edge_to: lambda _df: _df.groupby(self.eventstream.schema.user_id)[edge_from].shift(-1)})
            .dropna(subset=[edge_to])
            .groupby([edge_from, edge_to])
            .size()
            .index
        )
        bigrams = df.assign(**{edge_to: lambda _df: _df.groupby(self.group_col)[edge_from].shift(-1)}).dropna(
            subset=[edge_to]
        )
        abs_values = bigrams.groupby([edge_from, edge_to])[self.weight_col].nunique()
        if self.weight_col != self.eventstream.schema.event_id:
            abs_values = abs_values.reindex(possible_transitions)
        edgelist = abs_values
        # denumerator_full = total number of transitions/users/sessions
        if norm_type == "full":
            if self.weight_col != "event_id":
                denumerator_full = bigrams[self.weight_col].nunique()
            else:
                denumerator_full = len(bigrams)
            edgelist = abs_values / denumerator_full
        # denumerator_node = total number of transitions/users/sessions that started with edge_from event
        if norm_type == "node":
            denumerator_node = bigrams.groupby([edge_from])[self.weight_col].nunique()
            edgelist = abs_values / denumerator_node
        if self.weight_col not in [self.eventstream.schema.event_id, self.eventstream.schema.user_id]:
            edgelist = edgelist.fillna(0)

            if norm_type is None:
                edgelist = edgelist.astype(int)
        edgelist = edgelist.reset_index()
        return edgelist

    def update_threshold(self, edges_threshold: ThresholdValueMap | None = None) -> None:
        # filter edges by threshold
        if edges_threshold:

            def is_out_of_threshold(row: pd.Series, column_name: str, threshold: ThresholdValue) -> bool:
                value = row[column_name]
                return not threshold["min"] <= value <= threshold["max"]

            self.edgelist_df[IS_OUT_OF_THRESHOLD_COL] = False

            for threshold_col, threshold in edges_threshold.items():
                if threshold_col in self.edgelist_df.columns:
                    self.edgelist_df[IS_OUT_OF_THRESHOLD_COL] = (
                        self.edgelist_df.apply(is_out_of_threshold, axis=1, args=(threshold_col, threshold))
                        | self.edgelist_df[IS_OUT_OF_THRESHOLD_COL]
                    )

    def get_filtered_edges(self) -> pd.DataFrame:
        filtered_edges = self.edgelist_df[~self.edgelist_df[IS_OUT_OF_THRESHOLD_COL]]
        filtered_edges = filtered_edges.drop(columns=[IS_OUT_OF_THRESHOLD_COL])
        return filtered_edges

    def get_min_max(self) -> ThresholdValueMap:
        result: ThresholdValueMap = {}

        if self.weight_cols is None:  # type: ignore
            return result

        if self.edgelist_df is None:  # type: ignore
            return result

        for weight_col in self.weight_cols:
            min = float(self.edgelist_df[weight_col].min())
            max = float(self.edgelist_df[weight_col].max())

            result[weight_col] = {"min": min, "max": max}

        return result

    def get_threshold_min_max(self) -> ThresholdValueMap:
        result: ThresholdValueMap = {}

        for weight_col in self.weight_cols:
            min: float | int = self.edgelist_df[weight_col].min()
            max: float | int = self.edgelist_df[weight_col].max()
            result[weight_col] = {
                "min": min,
                "max": max,
            }

        return result

    def fit_threshold(self, threshold: ThresholdValueMap, prev_min_max: ThresholdValueMap) -> ThresholdValueMap:
        new_threshold = threshold.copy()
        threshold_min_max = self.get_threshold_min_max()

        for weight_col, value in new_threshold.items():
            if weight_col not in prev_min_max:
                continue
            if weight_col not in threshold_min_max:
                continue

            if value["max"] == prev_min_max[weight_col]["max"]:
                new_threshold[weight_col]["max"] = threshold_min_max[weight_col]["max"]

        return new_threshold

    def copy(self) -> Edgelist:
        el = Edgelist(eventstream=self.eventstream, edgelist_df=self.edgelist_df.copy())

        if self.weight_cols is not None:  # type: ignore
            el.weight_cols = self.weight_cols.copy()

        if self.weight_col:
            el.weight_col = self.weight_col

        return el



from __future__ import annotations

from typing import Literal, TypedDict, Union

from typing_extensions import NotRequired

NormalizationId = Literal["none", "full", "node"]

NormalizationType = Literal["absolute", "relative"]


class Normalization(TypedDict):
    id: NormalizationId
    name: str
    type: NormalizationType
    description: NotRequired[str]


NormType = Union[Literal["full", "node", "none"], None]



from __future__ import annotations

import uuid
from typing import Any, Callable, Optional, Dict

from traitlets import Unicode

from retentioneering.exceptions.server import ServerNotFoundActionError, ServerErrorWithResponse
from retentioneering.utils.singleton import Singleton
from .event_bridge import EventBridge


class Action:
    def __init__(self, method: str, callback: Callable):
        self.method = method
        self.callback = callback


class JupyterServer:
    def __init__(self, pk: Optional[str] = None, bridge: Optional[EventBridge] = None):
        self.pk = pk or self._make_id()
        self.actions: Dict[str, Action] = {}
        self._bridge = bridge

    def _make_id(self) -> str:
        return str(uuid.uuid4())

    def register_action(self, method: str, callback: Callable) -> None:
        self.actions[method] = Action(method, callback)

    def _find_action(self, method: str) -> Action | None:
        return self.actions.get(method)

    def dispatch_method(self, method: str, payload: dict) -> Any:
        action = self._find_action(method)
        if action is not None:
            return action.callback(payload)
        raise ServerNotFoundActionError("method not found!", method=method)


from typing import List, Optional

import pandas as pd
import plotly
import plotly.express as px
import plotly.graph_objects as go

from retentioneering.eventstream.segments import get_all_segments
from retentioneering.eventstream.types import EventstreamType


class SegmentProjection:
    def __init__(self, eventstream: EventstreamType):
        self.__eventstream = eventstream
        self.segment_cols: List[str]
        self.proj: pd.DataFrame

    def fit(self, proj: pd.DataFrame, segments: Optional[List[str]] = None) -> None:
        path_segments = get_all_segments(self.__eventstream, resolve_collision="majority")

        if segments is not None:
            path_segments = path_segments[segments]

        proj = proj.join(path_segments, how="left").reset_index()

        self.proj = proj
        self.segment_cols = path_segments.columns.tolist()

    def plot(self) -> plotly.graph_objects.Figure:
        path_id_col = self.__eventstream.schema.user_id
        x_col = self.proj.columns[1]
        y_col = self.proj.columns[2]

        segment_values_nunique = self.proj[self.segment_cols].nunique()
        toggles_off = [False] * (segment_values_nunique.sum() + 1)

        default_toggles = toggles_off.copy()
        default_toggles[0] = True

        dropdowns = [
            {"label": "None", "method": "update", "args": [{"visible": default_toggles}, {"title": "2D-projection"}]}
        ]

        plot_data = px.scatter(self.proj, x=x_col, y=y_col, hover_data=[path_id_col]).data

        trace_index = 1
        for segment_name in self.segment_cols:
            data = px.scatter(self.proj, x=x_col, y=y_col, color=segment_name, hover_data=[path_id_col, segment_name])
            data.for_each_trace(lambda trace: trace.update(visible=False))
            plot_data += data.data

            segment_toggles = toggles_off.copy()
            for _ in range(segment_values_nunique[segment_name]):
                segment_toggles[trace_index] = True
                trace_index += 1

            dropdowns.append(
                {
                    "label": segment_name,
                    "method": "update",
                    "args": [{"visible": segment_toggles}, {"title": f"2D-projection<br>(by {segment_name})"}],
                }
            )

        fig = go.Figure(data=plot_data)

        fig.update_layout(
            width=900,
            height=800,
            title="2D-projection",
            title_x=0.5,
            updatemenus=[{"buttons": dropdowns, "x": 1.2, "y": 1.06}],
            annotations=[
                {"text": "Segment", "x": 1.11, "xref": "paper", "y": 1.1, "yref": "paper", "showarrow": False}
            ],
        )

        return fig

    @property
    def values(self) -> pd.DataFrame:
        return self.proj



from __future__ import annotations

import uuid
import warnings
from typing import Final, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.common.constants import DATETIME_UNITS
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ListOfString, ReteTimeWidget


class SplitSessionsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.SplitSessions` class.

    """

    timeout: Optional[Tuple[float, DATETIME_UNITS]]
    delimiter_events: Optional[List[str]]
    delimiter_col: Optional[str]
    mark_truncated: bool = False
    session_col: str = "session_id"

    _widgets = {"timeout": ReteTimeWidget(), "delimiter_events": ListOfString()}


@docstrings.get_sections(base="SplitSessions")  # type: ignore
class SplitSessions(DataProcessor):
    """
    Create new synthetic events, that divide users' paths on sessions:
    ``session_start`` (or ``session_start_cropped``) and ``session_end`` (or ``session_end_cropped``).
    Also create a new column that contains session number for each event in input eventstream.
    Session number will take the form: ``{user_id}_{session_number through one user path}``.

    Parameters
    ----------
    timeout : Tuple(float, :numpy_link:`DATETIME_UNITS<>`), optional
        Threshold value and its unit of measure.
        ``session_start`` and ``session_end`` events are always placed before the first and after the last event
        in each user's path.
        Because user can have more than one session, it calculates timedelta between every two consecutive events in
        each user's path.
        If the calculated timedelta is more than selected timeout,
        new synthetic events - ``session_start`` and ``session_end`` are created inside the user path,
        marking session starting and ending points.
    delimiter_events : list of str, optional
        Delimiters define special events in the eventstream that indicate the start and the end of a session.

        - If a single delimiter is defined, it is associated with the session start and the end simultaneously.
          Delimiting events will be replaced with "session_start" event.
        - If a list of two delimiters is defined, the first and second events are associated with session start
          and end correspondingly. Delimiting events will be replaced with "session_start" and "session_end" events.

    delimiter_col : list, optional
        Determines a column that already contains custom session identifiers.

    mark_truncated : bool, default False
        Works with ``timeout`` argument only. If ``True`` - calculates timedelta between:

        - first event in each user's path and first event in the whole eventstream.
        - last event in each user's path and last event in the whole eventstream.

        For users with timedelta less than selected ``timeout``,
        a new synthetic event - ``session_start_cropped`` or ``session_end_cropped`` will be added.
    session_col : str, default "session_id"
        The name of the ``session_col``.

    Returns
    -------
    Eventstream
        ``Eventstream`` with new synthetic events and ``session_col``.

        +-----------------------------+----------------------------+-----------------+
        | **event_name**              | **event_type**             | **timestamp**   |
        +-----------------------------+----------------------------+-----------------+
        | session_start               | session_start              | first_event     |
        +-----------------------------+----------------------------+-----------------+
        | session_end                 | session_end                | last_event      |
        +-----------------------------+----------------------------+-----------------+
        | session_start_cropped       | session_start_cropped      | first_event     |
        +-----------------------------+----------------------------+-----------------+
        | session_end_cropped         | session_end_cropped        | last_event      |
        +-----------------------------+----------------------------+-----------------+

        If the delta between timestamps of two consecutive events
        (raw_event_n and raw_event_n+1) is greater than the selected ``timeout``
        the user will have more than one session:

        +--------------+-------------------+------------------+-------------------+------------------+
        |  **user_id** | **event_name**    | **event_type**   | **timestamp**     | **session_col**  |
        +--------------+-------------------+------------------+-------------------+------------------+
        |     1        | session_start     | session_start    | first_event       |     1_0          |
        +--------------+-------------------+------------------+-------------------+------------------+
        |     1        | session_end       | session_end      | raw_event_n       |     1_0          |
        +--------------+-------------------+------------------+-------------------+------------------+
        |     1        | session_start     | session_start    | raw_event_n+1     |     1_1          |
        +--------------+-------------------+------------------+-------------------+------------------+
        |     1        | session_end       | session_end      | last_event        |     1_1          |
        +--------------+-------------------+------------------+-------------------+------------------+

    Examples
    --------
    Splitting with a single delimiting event.

    .. code-block:: python

        df = pd.DataFrame(
            [
                [111, "session_delimiter", "2023-01-01 00:00:00"],
                [111, "A", "2023-01-01 00:00:01"],
                [111, "B", "2023-01-01 00:00:02"],
                [111, "session_delimiter", "2023-01-01 00:00:04"],
                [111, "C", "2023-01-01 00:00:04"],
            ],
            columns=["user_id", "event", "timestamp"]
        )
        Eventstream(df)\\
            .split_sessions(delimiter_events=["session_delimiter"])\\
            .to_dataframe()\\
            .sort_values(["user_id", "event_index"])\\
            [["user_id", "event", "timestamp", "session_id"]]

           user_id          event           timestamp session_id
        0      111  session_start 2023-01-01 00:00:00      111_1
        1      111              A 2023-01-01 00:00:01      111_1
        2      111              B 2023-01-01 00:00:02      111_1
        3      111    session_end 2023-01-01 00:00:02      111_1
        4      111  session_start 2023-01-01 00:00:04      111_2
        5      111              C 2023-01-01 00:00:04      111_2
        6      111    session_end 2023-01-01 00:00:04      111_2

    Splitting with a couple of delimiters indicating session start and session end.

    .. code-block:: python

        df = pd.DataFrame(
            [
                [111, "custom_start", "2023-01-01 00:00:00"],
                [111, "A", "2023-01-01 00:00:01"],
                [111, "B", "2023-01-01 00:00:02"],
                [111, "custom_end", "2023-01-01 00:00:02"],
                [111, "custom_start", "2023-01-01 00:00:04"],
                [111, "C", "2023-01-01 00:00:04"],
                [111, "custom_end", "2023-01-01 00:00:04"]
            ],
            columns=["user_id", "event", "timestamp"]
        )
        stream = Eventstream(df)
        stream.split_sessions(delimiter_events=["custom_start", "custom_end"])\\
            .to_dataframe()\\
            .sort_values(["user_id", "event_index"])\\
            [["user_id", "event", "timestamp", "session_id"]]

           user_id          event           timestamp session_id
        0      111  session_start 2023-01-01 00:00:00      111_1
        1      111              A 2023-01-01 00:00:01      111_1
        2      111              B 2023-01-01 00:00:02      111_1
        3      111    session_end 2023-01-01 00:00:02      111_1
        4      111  session_start 2023-01-01 00:00:04      111_2
        5      111              C 2023-01-01 00:00:04      111_2
        6      111    session_end 2023-01-01 00:00:04      111_2

    Splitting by a 'delimiter_col'.

    .. code-block:: python

        df = pd.DataFrame(
            [
                [111, "A", "2023-01-01 00:00:01", "session_1"],
                [111, "B", "2023-01-01 00:00:02", "session_1"],
                [111, "C", "2023-01-01 00:00:03", "session_2"],
                [111, "D", "2023-01-01 00:00:04", "session_2"],
            ],
            columns=["user_id", "event", "timestamp", "custom_ses_id"]
        )
        raw_data_schema = {"custom_cols": [{"raw_data_col": "custom_ses_id", "custom_col": "custom_ses_id"}]}
        stream = Eventstream(df, raw_data_schema=raw_data_schema)
        stream.split_sessions(delimiter_col="custom_ses_id")\\
            .to_dataframe()\\
            .sort_values(["user_id", "event_index"])\\
            [["user_id", "event", "timestamp", "session_id", "custom_ses_id"]]

           user_id          event           timestamp session_id custom_ses_id
        0      111  session_start 2023-01-01 00:00:01      111_1     session_1
        1      111              A 2023-01-01 00:00:01      111_1     session_1
        2      111              B 2023-01-01 00:00:02      111_1     session_1
        3      111    session_end 2023-01-01 00:00:02      111_1     session_1
        4      111  session_start 2023-01-01 00:00:03      111_2     session_2
        5      111              C 2023-01-01 00:00:03      111_2     session_2
        6      111              D 2023-01-01 00:00:04      111_2     session_2
        7      111    session_end 2023-01-01 00:00:04      111_2     session_2

    See Also
    --------
    .TimedeltaHist : Plot the distribution of the time deltas between two events.
    .Eventstream.describe : Show general eventstream statistics.
    .Eventstream.describe_events : Show general eventstream events statistics.


    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.

    """

    params: SplitSessionsParams

    @time_performance(
        scope="split_sessions",
        event_name="init",
    )
    def __init__(self, params: SplitSessionsParams) -> None:
        super().__init__(params=params)

        self.IS_SESSION_START_COL: Final = "is_session_start"

    def __get_session_borders_mask_by_timeout(
        self, df: pd.DataFrame, schema: EventstreamSchemaType
    ) -> Tuple[pd.Series, pd.Series]:
        timeout, timeout_unit = self.params.timeout or (None, None)
        user_col = schema.user_id
        timestamp_col = schema.event_timestamp

        prev_timedelta = df[timestamp_col] - df.groupby(user_col)[timestamp_col].shift(1)
        next_timedelta = df.groupby(user_col)[timestamp_col].shift(-1) - df[timestamp_col]
        prev_timedelta /= np.timedelta64(1, timeout_unit)  # type: ignore
        next_timedelta /= np.timedelta64(1, timeout_unit)  # type: ignore

        session_starts_mask = (prev_timedelta > timeout) | (prev_timedelta.isnull())
        session_ends_mask = (next_timedelta > timeout) | (next_timedelta.isnull())

        return session_starts_mask, session_ends_mask

    def __get_session_borders_mask_by_delimiting_events(
        self, df: pd.DataFrame, schema: EventstreamSchemaType
    ) -> Tuple[pd.Series, pd.Series, str, Union[str, None]]:
        delimiter_events = self.params.delimiter_events or [""]
        event_col = schema.event_name
        starting_event, ending_event = None, None

        if len(delimiter_events) == 1:
            starting_event = delimiter_events[0]
            session_starts_mask = df[event_col] == starting_event
            session_ends_mask = (df[event_col] == starting_event).shift(-1).fillna(True)
        elif len(delimiter_events) == 2:
            starting_event, ending_event = delimiter_events
            session_starts_mask = df[event_col] == starting_event
            session_ends_mask = df[event_col] == ending_event
        else:
            raise ValueError("The length of the 'delimiter_events' list must be 1 or 2.")

        return session_starts_mask, session_ends_mask, starting_event, ending_event

    def __get_session_borders_mask_by_delimiting_col(
        self, df: pd.DataFrame, schema: EventstreamSchemaType
    ) -> Tuple[pd.Series, pd.Series]:
        delimiter_col = self.params.delimiter_col
        user_col = schema.user_id

        if delimiter_col == self.params.session_col:
            warnings.warn("'delimiter_col' and 'session_col' are the same. The output will re-write delimiter_col.")

        session_starts_mask = df.groupby(user_col)[delimiter_col].shift(1) != df[delimiter_col]  # type: ignore
        session_ends_mask = df.groupby(user_col)[delimiter_col].shift(-1) != df[delimiter_col]  # type: ignore

        return session_starts_mask, session_ends_mask

    @time_performance(  # type: ignore
        scope="split_sessions",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        id_col = schema.event_id
        user_col = schema.user_id
        time_col = schema.event_timestamp
        type_col = schema.event_type
        event_col = schema.event_name
        event_index_col = schema.event_index
        session_col = self.params.session_col
        # relate to 'delimiter_events' parameter
        starting_event, ending_event = None, None

        delimiter_params = [self.params.timeout, self.params.delimiter_events, self.params.delimiter_col]
        delimiter_params_count = sum([x is not None for x in delimiter_params])

        if delimiter_params_count > 1:
            text_message = "Only one of 'timeout' or 'delimiter_events' or 'delimiter_col' parameter must be defined."
            raise ValueError(text_message)
        if delimiter_params_count == 0:
            text_message = "Either one of 'timeout' or 'delimiter_events' or 'delimiter_col' parameter must be defined."
            raise ValueError(text_message)

        df = df.sort_values([user_col, event_index_col])

        if self.params.timeout:
            session_starts_mask, session_ends_mask = self.__get_session_borders_mask_by_timeout(df, schema)
        elif self.params.delimiter_events:
            (
                session_starts_mask,
                session_ends_mask,
                starting_event,
                ending_event,
            ) = self.__get_session_borders_mask_by_delimiting_events(df, schema)
        elif self.params.delimiter_col:
            session_starts_mask, session_ends_mask = self.__get_session_borders_mask_by_delimiting_col(df, schema)
        else:
            raise ValueError("Either 'timeout' or 'delimiter_events' or 'delimiter_col' parameter must be defined.")

        df[self.IS_SESSION_START_COL] = session_starts_mask
        df[session_col] = df.groupby(user_col)[self.IS_SESSION_START_COL].transform("cumsum")
        df[session_col] = df[user_col].astype(str) + "_" + df[session_col].astype(str)

        session_starts = df[session_starts_mask].copy()
        session_ends = df[session_ends_mask].copy()

        session_starts[event_col] = "session_start"
        session_starts[type_col] = "session_start"
        session_starts[id_col] = [uuid.uuid4() for x in range(len(session_starts))]

        session_ends[event_col] = "session_end"
        session_ends[type_col] = "session_end"
        session_ends[id_col] = [uuid.uuid4() for x in range(len(session_ends))]

        if self.params.timeout and self.params.mark_truncated:
            timeout, timeout_unit = self.params.timeout
            dataset_start = df[time_col].min()
            dataset_end = df[time_col].max()
            start_to_start = (session_starts[time_col] - dataset_start) / np.timedelta64(1, timeout_unit)
            end_to_end = (dataset_end - session_ends[time_col]) / np.timedelta64(1, timeout_unit)

            session_starts_truncated = session_starts[start_to_start < timeout]
            session_ends_truncated = session_ends[end_to_end < timeout]

            session_starts_truncated[event_col] = "session_start_cropped"
            session_starts_truncated[type_col] = "session_start_cropped"

            session_ends_truncated[event_col] = "session_end_cropped"
            session_ends_truncated[type_col] = "session_end_cropped"

            session_starts = pd.concat([session_starts, session_starts_truncated])
            session_ends = pd.concat([session_ends, session_ends_truncated])

        result = pd.concat([df, session_starts, session_ends])
        result = result.drop([self.IS_SESSION_START_COL], axis=1)
        df = df.drop([self.IS_SESSION_START_COL], axis=1)

        collect_data_performance(
            scope="split_sessions",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            not_hash_values=["timeout"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        if self.params.delimiter_events:
            result = result[result[event_col] != starting_event]
            if ending_event:
                result = result[result[event_col] != ending_event]

        return result



from __future__ import annotations

import typing
import uuid
from typing import Any, Type

import pandas as pd

from retentioneering.data_processor.registry import register_dataprocessor
from retentioneering.params_model import ParamsModel
from retentioneering.utils.classes import call_if_implemented

if typing.TYPE_CHECKING:
    from retentioneering.eventstream.types import EventstreamSchemaType, EventstreamType


class DataProcessor:
    params: ParamsModel

    @classmethod
    def __init_subclass__(cls: Type[DataProcessor], **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        register_dataprocessor(cls)

    def __init__(self, params: ParamsModel | Any) -> None:
        if not issubclass(type(params), ParamsModel):
            raise TypeError("params is not subclass of ParamsModel")

        self.params = params
        self.pk = uuid.uuid4()

    def __call__(self, params: ParamsModel) -> DataProcessor:
        DataProcessor.__init__(self, params=params)
        return self

    def _define_new_schema(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> EventstreamSchemaType:
        cols = schema.get_default_cols()
        data_cols = list(df.columns)

        new_schema = schema.copy()
        new_schema.custom_cols = []

        if lost_cols := set(cols).difference(set(data_cols)):
            raise ValueError(
                f"Dataprocessor error, the following columns are missing from the returned dataframe:  {', '.join(lost_cols)}"
            )

        if len(data_cols) > len(cols):
            new_custom_cols = list(set(data_cols).difference(cols))
            new_schema.custom_cols.extend(new_custom_cols)

        return new_schema

    def _get_new_data(self, eventstream: EventstreamType) -> EventstreamType:
        from retentioneering.eventstream.eventstream import Eventstream

        copy = eventstream.copy()
        curr_events = copy.to_dataframe(drop_segment_events=False)
        new_events = self.apply(df=curr_events, schema=copy.schema)
        new_schema = self._define_new_schema(df=new_events, schema=copy.schema)

        new_stream = Eventstream(
            raw_data=new_events,
            raw_data_schema=new_schema.to_raw_data_schema(event_id=True, event_index=True),
            schema=new_schema,  # type: ignore
            index_order=copy.index_order,
            add_start_end_events=False,
        )

        return new_stream

    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        raise NotImplementedError

    def export(self) -> dict[str, Any]:
        data: dict[str, Any] = {}
        widgets: dict[str, Any] = self.params.get_widgets()
        data["name"] = self.__class__.__name__
        data["pk"] = str(self.pk)
        data["schema"] = self.params.schema()
        data["widgets"] = widgets
        return data

    @classmethod
    def get_view(cls) -> dict[str, str | list | dict]:
        data: dict[str, str | list | dict] = dict()
        data["name"] = cls.__name__
        from retentioneering.params_model.registry import params_model_registry

        params_models = params_model_registry.get_registry()
        params_model_name = cls.__annotations__["params"]
        if type(params_model_name) is str:
            params = params_models[params_model_name]
        else:
            params = params_model_name
        view = params.get_widgets()
        view_data = []
        for key in view:
            view_data.append(view[key])

        data["params"] = view_data
        return data

    def to_dict(self) -> dict:
        data = {
            "values": self.params.dict(),
            "name": self.__class__.__name__,
        }
        return data

    def copy(self) -> DataProcessor:
        return self.__copy__()

    def __copy__(self) -> DataProcessor:
        return self.__class__(params=self.params.copy())



from __future__ import annotations

from typing import Callable, Literal, Union

import pandas as pd

METRIC_PREFIX_HAS = "has:"
METRIC_PREFIX_TIME_TO_EVENT = "time_to:"
CLUSTERING_METHODS = Literal["kmeans", "gmm", "hdbscan"]
# https://numpy.org/doc/stable/reference/arrays.datetime.html#datetime-units
DATETIME_UNITS = Literal["Y", "M", "W", "D", "h", "m", "s", "ms", "us", "μs", "ns", "ps", "fs", "as"]
DATETIME_UNITS_LIST = ["Y", "M", "W", "D", "h", "m", "s", "ms", "us", "μs", "ns", "ps", "fs", "as"]
FEATURE_TYPES = Literal["tfidf", "count", "frequency", "binary", "time", "time_fraction", "markov"]
NGRAM_SEP = "->"
PATH_METRIC_TYPES = Union[str, Callable, pd.NamedAgg]
PROJECTION_METHODS = Literal["tsne", "umap"]
SCALERS = Literal["minmax", "std"]
SEGMENT_METRICS_OVERVIEW_KINDS = Literal["heatmap", "bar"]
SEQUENCES_METRIC_TYPES = Literal["paths", "paths_share", "count", "count_share"]
SKLEARN_FEATURE_TYPES = Literal["count", "frequency", "tfidf", "binary"]



from __future__ import annotations

from typing import Any, Type


class RegistryValidationError(Exception):
    pass


class ReteRegistry:
    REGISTRY: list[dict[str, Any]] = []
    objects: str = "object"

    def __setitem__(self, key: str, value: Any) -> None:
        # @TODO: fix bug. Vladimir Makhanov
        if key not in self.REGISTRY:  # type: ignore
            self.REGISTRY.append({key: value})
        else:
            raise RegistryValidationError("%s <%s> already exists" % (self.objects, key))

    @classmethod
    def remove(cls: Type[ReteRegistry], value: Any) -> None:
        cls.REGISTRY.remove(value)

    @classmethod
    def get_registry(cls: Type[ReteRegistry]) -> list[dict[str, Any]]:
        return cls.REGISTRY

    @classmethod
    def get_export_registry(cls: Type[ReteRegistry]) -> list:
        data = []
        for processor in cls.REGISTRY:
            key = list(processor.keys())[0]
            data.append(processor[key])
        return data



from __future__ import annotations

from typing import List, Optional, Tuple

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.common.constants import DATETIME_UNITS
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class LabelLostUsersHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(
        scope="label_lost_users",
        event_name="helper",
        event_value="combine",
    )
    def label_lost_users(
        self: EventstreamType,
        timeout: Optional[Tuple[float, DATETIME_UNITS]] = None,
        lost_users_list: Optional[List[int]] = None,
    ) -> EventstreamType:
        """
        A method of ``Eventstream`` class that creates one
        of the synthetic events in each user's path: ``lost_user`` or ``absent_user`` .

        Parameters
        ----------
            %(LabelLostUsers.parameters)s

        Returns
        -------
        Eventstream
             Input ``eventstream`` with new synthetic events.



        """
        calling_params = {
            "timeout": timeout,
            "lost_users_list": lost_users_list,
        }
        not_hash_values = ["timeout"]

        # avoid circular import
        from retentioneering.data_processors_lib import (
            LabelLostUsers,
            LabelLostUsersParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(
            processor=LabelLostUsers(
                params=LabelLostUsersParams(timeout=timeout, lost_users_list=lost_users_list)  # type: ignore
            )
        )
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="label_lost_users",
            event_name="metadata",
            called_params=calling_params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



from __future__ import annotations

from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, Extra, create_model_from_typeddict
from typing_extensions import (  # required for pydantic and python < 3.9.2
    NotRequired,
    Required,
    TypedDict,
)


class NodeData(TypedDict, total=False):
    name: Required[str]
    pk: Required[str]
    description: NotRequired[str]
    processor: NotRequired[Dict]


class NodeLink(TypedDict):
    source: str
    target: str


class Payload(TypedDict):
    directed: bool
    nodes: list[NodeData]
    links: list[NodeLink]


class CombineHandlerPayload(TypedDict):
    node_pk: str


class FieldErrorDesc(TypedDict):
    field: str
    msg: str


class CreateNodeErrorDesc(TypedDict):
    type: Literal["node_error"]
    node_pk: str
    msg: Optional[str]
    fields_errors: List[FieldErrorDesc]


class PreprocessingGraphConfig(BaseModel):
    directed: bool
    nodes: List[NodeData]
    links: List[NodeLink]

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow



from __future__ import annotations

import itertools
from typing import Callable, Tuple

import matplotlib.colors as mcolors
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import colormaps, colors
from scipy.sparse import csc_matrix

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.common.constants import NGRAM_SEP, SEQUENCES_METRIC_TYPES
from retentioneering.eventstream.segments import _split_segment
from retentioneering.eventstream.types import (
    EventstreamType,
    SplitExpr,
    UserGroupsNamesType,
    UserGroupsType,
)


class Sequences:
    """
    A class that provides methods for patterns exploration.

    Parameters
    ----------
    eventstream : EventstreamType


    See Also
    --------
    .Eventstream.sequences : Call Sequences tool as an eventstream method.


    Notes
    -----
    See :doc:`Sequences user guide</user_guides/sequences>` for the details.
    """

    __eventstream: EventstreamType
    ngram_range: Tuple[int, int]
    groups: UserGroupsType | None
    path_id_col: str
    metrics: SEQUENCES_METRIC_TYPES | None
    final_columns_level_0: list[str]
    group_names: UserGroupsNamesType | None
    DEFAULT_GROUP_NAMES = ("group_1", "group_2")
    SEQUENCE_TYPE_GROUPS_COLUMN_NAME = ("sequence_type", "")
    SEQUENCE_TYPE_COLUMN_NAME = "sequence_type"
    RELATIVE_DELTA_COLUMN_NAME = "delta_rel"
    ABS_DELTA_COLUMN_NAME = "delta_abs"
    ORIGINAL_METRICS_LIST = ["paths", "paths_share", "count", "count_share", "avg_count"]

    _vec_data: pd.DataFrame

    @time_performance(
        scope="sequences",
        event_name="init",
    )
    def __init__(self, eventstream: EventstreamType):
        self._default_columns_level_0: list = []
        self._sample_column_name: str
        self._full_metrics_list: list = []
        self._samples_full = None
        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp
        self._sequence_df = pd.DataFrame()
        self._show_values_df = pd.DataFrame()
        self.groups = None

    def _calculate_metrics_df(self, vec_data: pd.DataFrame) -> pd.DataFrame:
        sequence_df = pd.concat(
            [vec_data.sum(), vec_data.astype(bool).sum(axis=0)], axis=1, keys=["count", self.path_id_col]
        ).astype(int)

        sequence_df["count_share"] = sequence_df["count"] / sequence_df["count"].sum()
        sequence_df[f"{self.path_id_col}_share"] = sequence_df[self.path_id_col] / len(vec_data)
        sequence_df["avg_count"] = sequence_df["count"] / sequence_df[self.path_id_col]
        sequence_df = sequence_df[self._full_metrics_list].fillna(0)

        return sequence_df

    def _add_support_set(self, vec_data: pd.DataFrame) -> pd.DataFrame:
        sparse_matrix = csc_matrix(vec_data.values)
        idx = vec_data.index
        samples_full = []
        for j, col in enumerate(vec_data.columns):
            non_zero_indices = sparse_matrix.indices[sparse_matrix.indptr[j] : sparse_matrix.indptr[j + 1]]
            samples_full.append([idx[non_zero_indices].tolist()])
        samples_full = pd.DataFrame(samples_full, index=vec_data.columns, columns=[self._sample_column_name])

        return samples_full

    @staticmethod
    def _sequence_type(x: str) -> str:
        temp = x.split(NGRAM_SEP)
        n = len(temp)
        n_unique = len(set(temp))
        if (n_unique > 1) and (temp[0] == temp[-1]):
            return "cycle"
        if (n_unique == 1) and (n > 1):
            return "loop"
        return "other"

    def _build_groups_df(
        self, vec_data: pd.DataFrame, groups: UserGroupsType, group_names: UserGroupsNamesType
    ) -> pd.DataFrame:
        vec_data["group"] = None
        vec_data.loc[groups[0], "group"] = group_names[0]
        vec_data.loc[groups[1], "group"] = group_names[1]
        group_mask = vec_data["group"].copy()
        vec_data.drop(columns="group", inplace=True)
        group_dfs = []
        samples_dfs = []
        for group in group_names:
            vec_data_group = vec_data[group_mask == group]
            sequence_df_group = self._calculate_metrics_df(vec_data_group)
            sequence_df_group.columns = pd.MultiIndex.from_product([sequence_df_group.columns, [group]])
            group_dfs.append(sequence_df_group)
            samples_full_group = self._add_support_set(vec_data=vec_data_group)
            samples_dfs.append(samples_full_group)

        sequence_df = pd.concat(group_dfs, axis=1)
        sequence_type = pd.Series(sequence_df.index).apply(lambda x: self._sequence_type(x))
        sequence_df[self.SEQUENCE_TYPE_GROUPS_COLUMN_NAME] = sequence_type.values

        samples_full = pd.concat(samples_dfs, axis=1)
        samples_columns = [(self._sample_column_name, group_names[0]), (self._sample_column_name, group_names[1])]
        samples_full.columns = pd.MultiIndex.from_tuples(samples_columns)

        for metric in self._full_metrics_list:
            sequence_df[(metric, self.ABS_DELTA_COLUMN_NAME)] = (
                sequence_df[(metric, group_names[0])] - sequence_df[(metric, group_names[1])]
            )
            sequence_df[(metric, self.RELATIVE_DELTA_COLUMN_NAME)] = (
                sequence_df[(metric, self.ABS_DELTA_COLUMN_NAME)] / sequence_df[(metric, group_names[1])]
            )

        final_columns = (
            list(itertools.product(self._full_metrics_list, self._default_columns_level_0))
            + [self.SEQUENCE_TYPE_GROUPS_COLUMN_NAME]
            + samples_columns
        )
        index_cols = pd.MultiIndex.from_tuples(final_columns)

        sequence_df = sequence_df.merge(samples_full, left_index=True, right_index=True)

        return sequence_df[index_cols]

    @time_performance(
        scope="sequences",
        event_name="fit",
    )
    def fit(
        self,
        ngram_range: Tuple[int, int] = (1, 1),
        groups: SplitExpr | None = None,
        group_names: UserGroupsNamesType | None = None,
        path_id_col: str | None = None,
    ) -> None:
        """
        Calculate statistics on n-grams found in eventstream.
        Calculated path_metrics:

        - ``paths``: The number of unique paths that contain each particular event sequence
          (calculated within the specified ``path_id_col``, or ``user_id`` by default).
        - ``paths_share``: The ratio of paths to the sum of paths.
        - ``count``: The number of occurrences of a particular sequence.
        - ``count_share``: The ratio of count to the sum of counts.
        - ``avg_count``: The average number of occurrences per path.

        Defined sequences types:

        - ``loop`` - if sequence length >= 2 and all the events are equal.
        - ``cycle`` - if sequence length >= 3 and start and end events are equal.
        - ``other`` - all other sequences.

        Parameters
        ----------
        ngram_range : Tuple(int, int)
            The lower and upper boundary of the range of n-values for different word n-grams to be
            extracted. For example, ngram_range=(1, 1) means only single events, (1, 2) means single events
            and bigrams.
        groups : SplitExpr, optional
            Can be specified to calculate statistics for n-grams group-wise and provide delta values.
            Must contain a tuple of two elements (g_1, g_2), where g_1 and g_2 are collections
            of path IDs (for the column specified in the ``path_id_col`` parameter).
        group_names : UserGroupsNamesType, optional
            Names for the selected groups g_1 and g_2, which will be shown in the final plot header.
        path_id_col : str, optional
            The column used for calculating 'paths' and 'paths_share' path_metrics.
            If not specified, the ``user_id`` from ``eventstream.schema``
            will be used. For example, it can be specified as ``session_id`` if ``eventstream``
            has such a ``custom_col``.


        Returns
        -------
        None

        Notes
        -----
        See the results of calculation using :py:func:`plot` method and the :py:func:`values` attribute.

        """
        called_params = {
            "ngram_range": ngram_range,
            "groups": groups,
            "group_names": group_names,
            "path_id_col": path_id_col,
        }
        not_hash_values = ["ngram_range"]

        self.ngram_range = ngram_range
        self.path_id_col = path_id_col or self.__eventstream.schema.user_id
        self._sample_column_name = f"{self.path_id_col}_sample"
        self._full_metrics_list = [self.path_id_col, f"{self.path_id_col}_share", "count", "count_share", "avg_count"]

        if groups:
            self.groups, _ = _split_segment(self.__eventstream, groups)

        if group_names:
            self.group_names = group_names

        else:
            self.group_names = self.DEFAULT_GROUP_NAMES

        vec_data = self.__eventstream.extract_features(
            feature_type="count", ngram_range=self.ngram_range, path_id_col=self.path_id_col
        )

        if self.groups:
            self._default_columns_level_0 = [
                self.group_names[0],
                self.group_names[1],
                self.ABS_DELTA_COLUMN_NAME,
                self.RELATIVE_DELTA_COLUMN_NAME,
            ]
            vec_data = vec_data[vec_data.index.isin(list(self.groups[0]) + list(self.groups[1]))]  # type: ignore

        if self.groups:
            self._sequence_df = self._build_groups_df(
                vec_data=vec_data, groups=self.groups, group_names=self.group_names
            )

        else:
            sequence_df = self._calculate_metrics_df(vec_data)

            sequence_type = pd.Series(sequence_df.index).apply(lambda x: self._sequence_type(x))
            sequence_df[self.SEQUENCE_TYPE_COLUMN_NAME] = sequence_type.values
            samples_full = self._add_support_set(vec_data=vec_data)
            self._sequence_df = sequence_df.join(samples_full[self._sample_column_name])

        self._sequence_df.index.name = "Sequence"
        self._show_values_df = self._sequence_df.copy()

        collect_data_performance(
            scope="sequences",
            event_name="metadata",
            called_params=called_params,
            not_hash_values=not_hash_values,
            performance_data={"shape": self._show_values_df.shape},
            eventstream_index=self.__eventstream._eventstream_index,
        )

    def _size_sample_ids(
        self, data: pd.DataFrame, sample_size: int | None, final_columns_level_0: list
    ) -> pd.DataFrame:
        if sample_size:
            displayed_columns = final_columns_level_0 + [self.SEQUENCE_TYPE_COLUMN_NAME] + [self._sample_column_name]
            data = data[displayed_columns].copy()

            def shuffle_and_slice(x: pd.Series, size: int = sample_size) -> pd.Series:  # type: ignore
                np.random.seed(42)
                np.random.shuffle(x)
                return x[:size]

            if self.groups:
                for group in self.group_names:  # type: ignore
                    data[(self._sample_column_name, group)] = data[(self._sample_column_name, group)].apply(
                        lambda x: shuffle_and_slice(x)
                    )
            else:
                data[self._sample_column_name] = data[self._sample_column_name].apply(lambda x: shuffle_and_slice(x))  # type: ignore

        else:
            displayed_columns = final_columns_level_0 + [self.SEQUENCE_TYPE_COLUMN_NAME]
            data = self._sequence_df[displayed_columns]

        return data

    @staticmethod
    def _create_plot(
        data: pd.DataFrame, heatmap_columns: list, precision: int | None
    ) -> pd.io.formats.style.Styler:  # type: ignore
        columns_intersection = set(data.columns) & set(heatmap_columns)
        if not columns_intersection:
            return

        heatmap_values = data.copy(deep=True)
        numeric_cols = data.select_dtypes(include="number").columns
        max_values = data[numeric_cols].replace(+np.inf, -np.inf).max()
        min_values = data[numeric_cols].replace(-np.inf, +np.inf).min()
        heatmap_values[numeric_cols] = heatmap_values[numeric_cols].replace(+np.inf, max_values)  # type: ignore
        heatmap_values[numeric_cols] = heatmap_values[numeric_cols].replace(-np.inf, min_values)  # type: ignore

        cm = sns.diverging_palette(h_neg=246, h_pos=27, s=99, l=68, as_cmap=True)

        def add_background(s: pd.Series, cmap: str = "PuBu") -> list[str]:  # type: ignore
            a = heatmap_values.loc[:, s.name].copy()  # type: ignore

            if a.min() >= 0:
                norm = mcolors.TwoSlopeNorm(vmin=-1, vcenter=0, vmax=a.max())

            elif a.max() <= 0:
                norm = mcolors.TwoSlopeNorm(vmin=a.min(), vcenter=0, vmax=1)

            else:
                norm = mcolors.TwoSlopeNorm(vmin=a.min(), vcenter=0.0, vmax=a.max())

            normed = norm(a.values)
            c = [colors.rgb2hex(x) for x in colormaps.get_cmap(cmap)(normed)]
            return ["background-color: %s" % color for color in c]

        plot_sequence_df_styler = data.style.apply(add_background, subset=heatmap_columns, cmap=cm).format(
            precision=precision, thousands=" "
        )
        return plot_sequence_df_styler

    def _check_params(
        self,
        plot_sequence_df: pd.DataFrame,
        metrics: SEQUENCES_METRIC_TYPES | None = None,
        threshold: tuple[str, float | int] | None = None,
        sorting: tuple[str | tuple, bool] | tuple[list[str | tuple], list[bool]] | None = None,
        heatmap_columns: str | list[str | tuple] | None = None,
    ) -> Tuple[list, tuple, tuple, list]:
        replacements = {"paths": self.path_id_col, "paths_share": f"{self.path_id_col}_share"}
        replacer = replacements.get

        final_columns_full, final_columns_level_0 = self._check_metrics(
            plot_sequence_df=plot_sequence_df, metrics=metrics, replacer=replacer
        )

        heatmap_columns_list: list
        heatmap_columns_list = self._define_heatmap_columns(
            heatmap_columns=heatmap_columns,
            replacer=replacer,
            final_columns_level_0=final_columns_level_0,
            final_columns_full=final_columns_full,
        )

        sorting_checked = self._define_sorting_columns(
            sorting=sorting, replacer=replacer, final_columns_full=final_columns_full
        )

        threshold_checked: Tuple[list | str | None, float | int | None]
        if threshold:
            threshold_checked = self._define_threshold_columns(
                threshold=threshold, replacer=replacer, final_columns_full=final_columns_full
            )
        else:
            threshold_checked = None, None

        return heatmap_columns_list, sorting_checked, threshold_checked, final_columns_level_0

    def _check_metrics(
        self, plot_sequence_df: pd.DataFrame, replacer: Callable, metrics: SEQUENCES_METRIC_TYPES | None = None
    ) -> Tuple[list, list]:
        if self.groups:
            final_columns_level_0 = [self.path_id_col]
        else:
            final_columns_level_0 = self._full_metrics_list

        if metrics:
            if isinstance(metrics, str):  # type: ignore
                final_metrics = [metrics]
            else:
                final_metrics = metrics
            if not set(final_metrics).issubset(set(self.ORIGINAL_METRICS_LIST)):
                unexpected_metrics = set(final_metrics).difference(set(self.ORIGINAL_METRICS_LIST))
                raise ValueError(
                    f"""Unexpected path_metrics value: {unexpected_metrics}. Should be a metric name or
                                 a list of metric names from {self.ORIGINAL_METRICS_LIST}."""
                )

            final_columns_level_0 = [replacer(n, n) for n in final_metrics]
        final_columns_full = list(plot_sequence_df[final_columns_level_0].columns)
        return final_columns_full, final_columns_level_0

    def _define_sorting_columns(
        self,
        sorting: tuple[str | tuple, bool] | tuple[list[str | tuple], list[bool]] | None,
        replacer: Callable,
        final_columns_full: list,
    ) -> Tuple[str | list | tuple, bool | list | tuple]:
        # sorting
        if sorting:
            error_message_text_sorting = f"""Should be tuple or list with two elements:
            (column_name, bool) or ([column_name], [list of bool]). Note that those two elements should be the
            same length. For example: ('count_share', False), ([('count_share', 'delta'),('paths_share', 'delta')],
            [False, True])."""
            sorting_columns: str | list | tuple
            sorting_direction: bool | list | tuple
            if not self.groups:
                if isinstance(sorting, (list, tuple)):  # type: ignore
                    if isinstance(sorting[0], str) and isinstance(sorting[1], bool):
                        sorting_columns = [sorting[0]]
                        sorting_direction = [sorting[1]]

                    elif len(sorting) == 2 and all(isinstance(elem, (list, tuple)) for elem in sorting):
                        sorting_columns = sorting[0]
                        sorting_direction = sorting[1]
                        if (
                            not all(isinstance(elem, str) for elem in sorting_columns)
                            or not all(isinstance(elem, bool) for elem in sorting_direction)  # type: ignore
                            or not len(sorting_columns) == len(sorting_direction)  # type: ignore
                        ):
                            raise TypeError(f"Unexpected sorting parameter type. {error_message_text_sorting}")
                    else:
                        raise TypeError(f"Unexpected sorting parameter type. {error_message_text_sorting}")
                else:
                    raise TypeError(f"Unexpected sorting parameter type. {error_message_text_sorting}")

                sorting_columns = [(replacer(col, col)) for col in sorting_columns]

            else:
                if isinstance(sorting, (list, tuple)):  # type: ignore
                    if isinstance(sorting[0], (list, tuple)) and isinstance(sorting[1], bool):
                        sorting_columns = [sorting[0]]
                        sorting_direction = [sorting[1]]

                    elif len(sorting) == 2 and all(isinstance(elem, (list, tuple)) for elem in sorting):
                        sorting_columns = sorting[0]
                        sorting_direction = sorting[1]
                        if len(sorting_columns) != len(sorting_direction):  # type: ignore
                            raise TypeError(f"Unexpected sorting parameter type. {error_message_text_sorting}")
                    else:
                        raise TypeError(f"Unexpected sorting parameter type. {error_message_text_sorting}")

                sorting_columns = [(replacer(col[0], col[0]), col[1]) for col in sorting_columns]

            if not set(sorting_columns).issubset(set(final_columns_full)):
                unexpected_sorting_columns = set(sorting_columns).difference(set(final_columns_full))
                raise ValueError(
                    f"""Unexpected sorting_columns name: {unexpected_sorting_columns}.
                                     {error_message_text_sorting}"""
                )
        else:
            if self.groups:
                sorting_columns = final_columns_full[3]
                sorting_direction = False
            else:
                sorting_columns = final_columns_full[0]
                sorting_direction = False

        return sorting_columns, sorting_direction

    def _define_threshold_columns(
        self, threshold: tuple[str | list, float | int], replacer: Callable, final_columns_full: list
    ) -> Tuple[list | str | None, float | int | None]:
        threshold_column, threshold_value = threshold
        error_message_text_thresh = f"""Should be a tuple or list with column_name as the first element and int or
                                        float values as the second.
                                        For example: ('count_share', 0.5), (('count_share', 'delta'), 0.5)."""
        if not isinstance(threshold_value, (int, float)):  # type: ignore
            raise TypeError(f"Unexpected threshold parameter type. {error_message_text_thresh}")

        if not self.groups:
            if not isinstance(threshold_column, str):  # type: ignore
                raise TypeError(f"Unexpected threshold parameter type. {error_message_text_thresh}")
            else:
                threshold_column = replacer(threshold_column, threshold_column)

        else:
            if not isinstance(threshold_column, (list, tuple)):
                raise TypeError(f"Unexpected threshold parameter type. {error_message_text_thresh}")
            elif not len(threshold_column) == 2 and all(isinstance(elem, str) for elem in threshold_column):  # type: ignore
                raise TypeError(f"Unexpected threshold parameter type. {error_message_text_thresh}")
            else:
                threshold_column = (replacer(threshold_column[0], threshold_column[0]), threshold_column[1])  # type: ignore

        if threshold_column not in final_columns_full:
            raise ValueError(
                f"""Unexpected sorting_columns name: {threshold_column}.
                                                 {error_message_text_thresh}"""
            )

        return threshold_column, threshold_value  # type: ignore

    def _define_heatmap_columns(
        self,
        heatmap_columns: str | list[str | tuple] | None,
        replacer: Callable,
        final_columns_level_0: list,
        final_columns_full: list,
    ) -> list:
        error_message_text_heatmap = f"""Should be a column name or a list
                                        of column names. For example: ['count_share', 'delta'],
                                        [('count_share', 'delta'),('paths_share', 'delta')]."""
        if heatmap_columns:
            if not self.groups:
                if isinstance(heatmap_columns, str):
                    heatmap_columns = [heatmap_columns]
                elif isinstance(heatmap_columns, (list, tuple)):  # type: ignore
                    if not all(isinstance(col, str) for col in heatmap_columns):
                        raise TypeError(f"Unexpected heatmap_columns type. {error_message_text_heatmap}")
                else:
                    raise TypeError(f"Unexpected heatmap_columns type. {error_message_text_heatmap}")

                heatmap_columns = [replacer(col, col) for col in heatmap_columns]

            else:
                if isinstance(heatmap_columns, (list, tuple)):
                    if len(heatmap_columns) == 2 and all(isinstance(col, str) for col in heatmap_columns):
                        heatmap_columns = [heatmap_columns]  # type: ignore
                    elif not all(isinstance(col, (list, tuple)) for col in heatmap_columns):
                        raise TypeError(f"Unexpected heatmap_columns type. {error_message_text_heatmap}")

                    heatmap_columns = [(replacer(col[0], col[0]), col[1]) for col in heatmap_columns]  # type: ignore

            if not set(heatmap_columns).issubset(set(final_columns_full)):
                unexpected_heatmap_columns = set(heatmap_columns).difference(set(final_columns_full))
                raise ValueError(
                    f"""Unexpected heatmap_columns name: {unexpected_heatmap_columns}.
                                            {error_message_text_heatmap}"""
                )

        else:
            if self.groups:
                heatmap_columns = list(itertools.product(final_columns_level_0, [self.RELATIVE_DELTA_COLUMN_NAME]))
            else:
                heatmap_columns = [f"{self.path_id_col}_share"]
        return heatmap_columns  # type: ignore

    @time_performance(
        scope="sequences",
        event_name="plot",
    )
    def plot(
        self,
        metrics: SEQUENCES_METRIC_TYPES | None = None,
        threshold: tuple[str, float | int] | None = None,
        sorting: tuple[str | tuple, bool] | tuple[list[str | tuple], list[bool]] | None = None,
        heatmap_cols: str | list[str | tuple] | None = None,
        sample_size: int | None = 1,
        precision: int = 2,
    ) -> pd.io.formats.style.Styler:  # type: ignore
        """
        Parameters
        ----------
        metrics : {'paths', 'paths_share', 'count', 'count_share'}, optional
            Specify the path_metrics to be displayed in the plot.

            - If groups are specified, by default, only the 'paths' metric will be plotted within each group,
              along with both deltas (relative and absolute).
            - If ``groups=None``, all four path_metrics will be shown by default.

        threshold : tuple[str, float | int], optional
            Used to filter out infrequent sequences based on the specified metric.

            - Example without groups: ('paths', 1200)
            - Example with groups: (('user_id', 'group_1'), 1200)

            Only rows with values greater than or equal to 1200 in the specified column will be displayed.

        sorting : Tuple(str or list of str, bool or list of bool) or None, default None
            - The first element in the tuple: Column name or list of names for sorting.
            - The second element: The sorting order (ascending vs. descending). Specify a list for multiple
              sorting orders.
              If a list of bools is provided, it must match the length of the sorting columns.

        heatmap_cols : str or list of str or list of tuples or None
            Specifies columns to be represented in the heatmap as follows:

            - The heatmap range is calculated column-wise.
            - For columns containing negative values, the palette will be divergent (blue - orange with white as zero).
            - For columns with only positive values, the palette will be orange.
            - For columns with only negative values, the palette will be blue.

            Default values

        sample_size : int or None, default 1
            Number of ID samples to display.
        precision : int, default 2
            Number of decimal digits to show as fractions in the heatmap.

        Returns
        -------
        pd.io.formats.style.Styler
            Styled pd.Dataframe object.

        """
        called_params = {
            "path_metrics": metrics,
            "threshold": threshold,
            "sorting": sorting,
            "heatmap_cols": heatmap_cols,
            "sample_size": sample_size,
            "precision": precision,
        }

        not_hash_values = ["path_metrics"]

        plot_sequence_df = self._sequence_df.copy()
        heatmap_columns_list, sorting_checked, threshold_checked, final_columns_level_0 = self._check_params(
            plot_sequence_df, metrics, threshold, sorting, heatmap_cols
        )

        sorting_columns, sorting_direction = sorting_checked
        threshold_column, threshold_value = threshold_checked
        plot_sequence_df = self._size_sample_ids(
            data=plot_sequence_df, sample_size=sample_size, final_columns_level_0=final_columns_level_0
        )

        plot_sequence_df = plot_sequence_df.sort_values(by=sorting_columns, ascending=sorting_direction)
        if threshold:
            plot_sequence_df = plot_sequence_df[plot_sequence_df[threshold_column] >= abs(threshold_value)]

        if precision:
            plot_sequence_df = plot_sequence_df.round(precision)

        plot_sequence_df_styler = self._create_plot(
            data=plot_sequence_df, heatmap_columns=heatmap_columns_list, precision=precision
        )

        self._show_values_df = plot_sequence_df
        collect_data_performance(
            scope="sequences",
            event_name="metadata",
            called_params=called_params,
            not_hash_values=not_hash_values,
            performance_data={"shape": plot_sequence_df.shape},
            eventstream_index=self.__eventstream._eventstream_index,
        )

        return plot_sequence_df_styler

    @property
    @time_performance(
        scope="sequences",
        event_name="values",
    )
    def values(self) -> pd.DataFrame:
        """
        Returns a pd.DataFrame representing the fitted or plotted Sequences table.
        Should be used after :py:func:`fit` or :py:func:`plot`.


        Returns
        -------
        pd.DataFrame



        """
        return self._show_values_df

    @property
    @time_performance(
        scope="sequences",
        event_name="params",
    )
    def params(self) -> dict[str, tuple | None | str | UserGroupsType | UserGroupsNamesType]:
        """
        Returns the parameters used for the last fitting.
        Should be used after :py:func:`fit`.

        """
        return {
            "ngram_range": self.ngram_range,
            "groups": self.groups,
            "group_names": self.group_names,
            "path_id": self.path_id_col,
        }



from __future__ import annotations

import json
import threading
from dataclasses import asdict

import requests
from requests.adapters import HTTPAdapter

from retentioneering.backend.tracker.connector import ConnectorProtocol
from retentioneering.backend.tracker.tracking_info import TrackingInfo


class TrackerMainConnector(ConnectorProtocol):
    url: str
    source: str
    session: requests.Session

    def __init__(self) -> None:
        super().__init__()
        self.url = "https://ftrack-1.server.retentioneering.com/endpoint/event"
        self.source = "rete_tools_backend"
        adapter = HTTPAdapter(pool_connections=30, pool_maxsize=30)
        self.session = requests.Session()
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)

    def _post(self, data: dict) -> None:
        # run _post_job in a separate thread using multi-threading
        try:
            post_thread = threading.Thread(target=self._post_job, args=(data,))
            post_thread.start()
        except Exception:
            # supress any exceptions in tracking
            # @TODO: store exceptions to tracking base or sentry. Vladimir Makhanov
            pass

    def _post_job(self, data: dict) -> requests.Response | None:
        try:
            data["source"] = self.source
            with self.session.post(self.url, data=json.dumps(data)) as response:
                return response
        except Exception:
            # supress any exceptions in tracking. Vladimir Makhanov
            return None

    def _prepare_data(self, data: TrackingInfo) -> dict:
        prepared_data = asdict(data)
        prepared_data["source"] = self.source
        prepared_data["params"] = json.dumps(prepared_data["params"])
        del prepared_data["event_time"]
        return prepared_data

    def send_message(self, data: TrackingInfo) -> None:
        prepared_data = self._prepare_data(data=data)
        self._post(prepared_data)



from inspect import signature
from typing import Any, Callable, Optional

import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ReteFunction

EventstreamFilter = Callable[[pd.DataFrame, Optional[EventstreamSchemaType]], Any]


class GroupEventsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.GroupEvents` class.
    """

    event_name: str
    func: EventstreamFilter
    event_type: Optional[str] = "group_alias"

    _widgets = {
        "func": ReteFunction(),
    }


@docstrings.get_sections(base="GroupEvents")  # type: ignore
class GroupEvents(DataProcessor):
    """
    Filter the specified events from the input ``eventstream`` and create
    new synthetic events, with names based on the old events' names.

    Parameters
    ----------
    event_name : str
        Name of the created event.
    func : Callable[[DataFrame, Optional[EventstreamSchema]], Any]
        Custom function that returns boolean mask with the same length as input eventstream.

        - If ``True`` - events will be grouped.
        - If ``False`` - events will be remained.
    event_type : str, default "group_alias"
        Event_type name for the grouped events.
        If custom event_type is created, it should be added to the ``DEFAULT_INDEX_ORDER``.

    Returns
    -------
    Eventstream
        ``Eventstream`` with:

         - new synthetic events with ``group_alias`` or custom type`

        +-----------------+----------------+-------------------+
        | **event_name**  | **event_type** | **timestamp**     |
        +-----------------+----------------+-------------------+
        | raw_event_name  | raw            | raw_event         |
        +-----------------+----------------+-------------------+
        | new_event_name  | group_alias    | raw_event         |
        +-----------------+----------------+-------------------+

    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.
    """

    params: GroupEventsParams

    @time_performance(
        scope="group_events",
        event_name="init",
    )
    def __init__(self, params: GroupEventsParams) -> None:
        super().__init__(params=params)

    @time_performance(
        scope="group_events",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        event_name = self.params.event_name
        func: Callable = self.params.func
        event_type = self.params.event_type

        parent_info = {
            "shape": df.shape,
            "hash": hash_dataframe(df),
        }

        expected_args_count = len(signature(func).parameters)
        if expected_args_count == 1:
            mask = func(df)  # type: ignore
        else:
            mask = func(df, schema)

        with pd.option_context("mode.chained_assignment", None):
            if event_type is not None:
                df.loc[mask, schema.event_type] = event_type
            df.loc[mask, schema.event_name] = event_name

        collect_data_performance(
            scope="group_events",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            performance_data={
                "parent": parent_info,
                "child": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
            },
        )

        return df



import numpy as np
import pandas as pd

from retentioneering.eventstream.types import EventstreamSchemaType


class EndedEventsMixin:
    def __init__(self) -> None:
        pass

    @staticmethod
    def _add_ended_events(data: pd.DataFrame, schema: EventstreamSchemaType, weight_col: str) -> pd.DataFrame:
        """
        Adds artificial ``ENDED`` event in the end of a path. If a path already
        contains ``path_end`` event, it will be replaced with ``ENDED`` event.
        Otherwise, ``ENDED`` event will be placed into the end of the path.
        Path is identified by the weight_col parameter which can be user_id, session_id, etc.
        """
        data[schema.event_name] = data[schema.event_name].str.replace("path_end", "ENDED")
        ids_with_ended = data[data[schema.event_name] == "ENDED"][weight_col].unique()

        paths_with_ended = data[data[weight_col].isin(ids_with_ended)]
        paths_without_ended = data[~data[weight_col].isin(ids_with_ended)]

        additional_ended_events = (
            paths_without_ended.groupby(weight_col, as_index=False)
            .last()
            .assign(
                **{
                    schema.event_name: "ENDED",
                    schema.event_index: lambda df_: df_[schema.event_index]
                    + (np.where(df_[schema.event_name] == "ENDED", 0.5, 0)),
                }
            )
        )

        new_data = pd.concat([paths_with_ended, paths_without_ended, additional_ended_events])
        return new_data



from typing import Any, List


def call_if_implemented(obj: Any, method_name: str, args: List[Any]) -> Any:
    method = getattr(obj, method_name, None)
    if method is None or not callable(method):
        return None
    try:
        return method(*args)
    except NotImplementedError:
        return None



from __future__ import annotations

from typing import Any, Dict, Optional

from IPython.core.display import display

from retentioneering.backend import JupyterServer
from retentioneering.exceptions.server import ServerErrorWithResponse
from retentioneering.utils.singleton import Singleton

from .event_bridge import EventBridge


class ServerManager:
    __metaclass__ = Singleton

    def __init__(self) -> None:
        self._servers: Dict[str, JupyterServer] = {}
        self._bridge: Optional[EventBridge] = None
        self._listener_created = False

        self._create_main_listener()
        display(self._bridge)

    def _find_server(self, server_id: str) -> JupyterServer | None:
        return self._servers.get(server_id)

    def _on_bridge_message(self, content: Dict[str, Any]) -> None:
        if self._bridge is None:
            return

        server_id = content.get("server_id")
        request_id = content.get("request_id")
        request_type = content.get("request_type")
        method = content.get("method")
        payload = content.get("payload", {})

        if server_id is None or request_id is None or (request_type != "handshake" and method is None):
            self._bridge.send(
                {
                    "success": False,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": "InvalidRequest",
                }
            )
            return

        # handshake
        if request_type == "handshake":
            self._bridge.send(
                {
                    "success": True,
                    "server_id": server_id,
                    "request_id": request_id,
                    "request_type": "handshake",
                }
            )
            return

        if method is None:
            return

        server = self._find_server(server_id)

        if server is None:
            self._bridge.send(
                {
                    "success": False,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": "ServerNotFound",
                }
            )
            return

        try:
            result = server.dispatch_method(method=method, payload=payload)
            self._bridge.send(
                {
                    "success": True,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": result,
                }
            )
        except ServerErrorWithResponse as err:
            self._bridge.send(
                {
                    "success": False,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": err.dict(),
                }
            )
        except Exception as err:
            wrapped = ServerErrorWithResponse(message=str(err), type="unexpected_error")
            self._bridge.send(
                {
                    "success": False,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": wrapped.dict(),
                }
            )

    def _create_main_listener(self) -> None:
        if self._listener_created:
            return
        bridge = EventBridge()

        def _recv(_: Any, content: Dict[str, Any], buffers: Any) -> None:
            self._on_bridge_message(content)

        bridge.on_msg(_recv)
        self._bridge = bridge
        self._listener_created = True

    def create_server(self, pk: Optional[str] = None) -> JupyterServer:
        server = JupyterServer(pk=pk, bridge=self._bridge)
        self._servers[server.pk] = server
        return server

    @property
    def widget(self) -> EventBridge | None:
        return self._bridge



from __future__ import annotations


class Singleton:
    _instances = {}  # type: ignore

    def __new__(cls, *args, **kwargs):  # type: ignore
        if cls._instances.get(cls, None) is None:
            cls._instances[cls] = super().__new__(cls)

        return Singleton._instances[cls]



from __future__ import annotations

import platform
from dataclasses import dataclass
from datetime import datetime
from typing import Any

from retentioneering import __version__
from retentioneering.backend import counter


@dataclass
class TrackingInfo:
    user_id: str
    event_custom_name: str
    event_name: str
    event_value: str
    params: dict[str, Any]
    scope: str
    event_time: datetime
    jupyter_kernel_id: str
    colab: int
    # marketing_session_type_3_id: str # @FIXME: what is this? Vladimir Makhanov.

    event_date_local: str = ""
    # event_date_moscow: str    # @TODO: implement. Vladimir Makhanov
    event_day_week: int = 0
    event_timestamp: int = 0
    event_timestamp_ms: int = 0
    source: str = "rete_tools_backend"
    version: str = str(__version__)
    os: str = platform.system()
    index: int = 0
    browser: str = ""
    eventstream_index: int | None = None
    parent_eventstream_index: int | None = None
    child_eventstream_index: int | None = None
    top_call_id: str = ""
    account_id: str = ""

    def __post_init__(self) -> None:
        event_timestamp = self.event_time.timestamp()
        tz_string = self.event_time.astimezone().tzname()
        local_printable_date = self.event_time.strftime("%m/%d/%Y %H:%M:%S")
        self.account_id = self.user_id
        self.event_date_local = f"{local_printable_date} GMT{tz_string}"
        self.event_day_week = self.event_time.weekday()
        self.event_timestamp = int(event_timestamp)
        self.event_timestamp_ms = int(event_timestamp * 1000)
        self.user_id = f"{self.user_id}|none|none|none"
        self.index = counter.get_event_index()



from __future__ import annotations

import uuid
from typing import Any, Optional, Type, Union

from retentioneering.data_processor import DataProcessor
from retentioneering.data_processor.registry import dataprocessor_registry
from retentioneering.eventstream.types import EventstreamType
from retentioneering.params_model.registry import params_model_registry


class BaseNode:
    processor: Optional[DataProcessor]
    events: Optional[EventstreamType]
    pk: str
    description: Optional[str]

    def __init__(self, **kwargs: Any) -> None:
        self.pk = str(uuid.uuid4())

    def __str__(self) -> str:
        data = {"name": self.__class__.__name__, "pk": self.pk}
        return str(data)

    __repr__ = __str__

    def export(self) -> dict:
        data: dict[str, Any] = {"name": self.__class__.__name__, "pk": self.pk}
        if self.description:
            data["description"] = self.description

        if processor := getattr(self, "processor", None):
            data["processor"] = processor.to_dict()
        return data

    def copy(self) -> BaseNode:
        return BaseNode()


class SourceNode(BaseNode):
    events: EventstreamType
    description: Optional[str]

    def __init__(self, source: EventstreamType, description: Optional[str] = None) -> None:
        super().__init__()
        self.events = source
        self.description = description

    def __copy__(self) -> SourceNode:
        return SourceNode(
            source=self.events.copy(),
            description=self.description,
        )

    def copy(self) -> SourceNode:
        return self.__copy__()


class EventsNode(BaseNode):
    """
    Class for regular nodes of a PreprocessingGraph.

    Notes
    -----
    See :doc:`Preprocessing user guide</user_guides/preprocessing>` for the details.

    See Also
    --------
    .PreprocessingGraph.add_node : Add a node to PreprocessingGraph.
    .PreprocessingGraph.combine : Run calculations of PreprocessingGraph.
    .MergeNode : Merging nodes of a PreprocessingGraph.

    """

    processor: DataProcessor
    events: Optional[EventstreamType]
    description: Optional[str]

    def __init__(self, processor: DataProcessor, description: Optional[str] = None) -> None:
        super().__init__()
        self.processor = processor
        self.events = None
        self.description = description

    def __copy__(self) -> EventsNode:
        return EventsNode(
            processor=self.processor.copy(),
            description=self.description,
        )

    def copy(self) -> EventsNode:
        return self.__copy__()


class MergeNode(BaseNode):
    """
    Class for merging nodes of a PreprocessingGraph.

    Notes
    -----
    See :doc:`Preprocessing user guide</user_guides/preprocessing>` for the details.

    See Also
    --------
    .PreprocessingGraph.add_node : Add a node to PreprocessingGraph.
    .PreprocessingGraph.combine : Run calculations of PreprocessingGraph.
    .EventsNode : Regular nodes of a PreprocessingGraph.

    """

    events: Optional[EventstreamType]
    description: Optional[str]

    def __init__(self, description: Optional[str] = None) -> None:
        super().__init__()
        self.events = None
        self.description = description

    def __copy__(self) -> MergeNode:
        return MergeNode(
            description=self.description,
        )

    def copy(self) -> MergeNode:
        return self.__copy__()


Node = Union[SourceNode, EventsNode, MergeNode]
nodes = {
    "MergeNode": MergeNode,
    "EventsNode": EventsNode,
    "SourceNode": SourceNode,
}


class NotFoundDataprocessor(Exception):
    pass


def build_node(
    source_stream: EventstreamType,
    pk: str,
    node_name: str,
    processor_name: str | None = None,
    processor_params: dict[str, Any] | None = None,
    descriptionn: Optional[str] = None,
) -> Node:
    _node = nodes[node_name]
    node_kwargs = {}

    if node_name == "SourceNode":
        node_kwargs["source"] = source_stream

    if not processor_params:
        processor_params = {}

    if processor_name and node_name == "EventsNode":
        _params_model_registry = params_model_registry.get_registry()
        _dataprocessor_registry = dataprocessor_registry.get_registry()

        _processor: Type[DataProcessor] = _dataprocessor_registry[processor_name]  # type: ignore
        params_name = _processor.__annotations__["params"]
        _params_model = _params_model_registry[params_name] if type(params_name) is str else params_name

        params_model = _params_model(**processor_params)

        processor: DataProcessor = _processor(params=params_model)
        node_kwargs["processor"] = processor  # type: ignore

    node = _node(**node_kwargs)
    node.pk = pk
    node.description = descriptionn
    return node



from __future__ import annotations

from typing import Optional, Tuple

import numpy as np
import pandas as pd
from pandas import DataFrame

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.common.constants import DATETIME_UNITS
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.types import EventstreamSchemaType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ReteTimeWidget


class LabelCroppedPathsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.LabelCroppedPaths` class.
    """

    left_cutoff: Optional[Tuple[float, DATETIME_UNITS]]
    right_cutoff: Optional[Tuple[float, DATETIME_UNITS]]

    _widgets = {
        "left_cutoff": ReteTimeWidget(),
        "right_cutoff": ReteTimeWidget(),
    }


@docstrings.get_sections(base="LabelCroppedPaths")  # type: ignore
class LabelCroppedPaths(DataProcessor):
    """
    Create new synthetic event(s) for each user based on the timeout threshold:
    ``cropped_left`` or ``cropped_right``

    Parameters
    ----------
    left_cutoff : Tuple(float, :numpy_link:`DATETIME_UNITS<>`), optional
        Threshold value with its unit of measure.
        The timedelta between the last event in each user's path and the first event in the whole eventstream
        is calculated. For users with the timedelta less than the selected ``left_cutoff``,
        the new ``cropped_left`` event is added.
    right_cutoff : Tuple(float, :numpy_link:`DATETIME_UNITS<>`), optional
        Threshold value with its unit of measure.
        The timedelta between the first event in each user's path and the last event in the whole eventstream
        is calculated. For users with timedelta less than the selected ``right_cutoff``,
        the new ``cropped_right`` event is added.

    Returns
    -------
    Eventstream
        Eventstream containing only the generated synthetic events, for users whose paths
        satisfy the specified cut-offs.

        +-------------------+-------------------+------------------+
        | **event_name**    | **event_type**    |  **timestamp**   |
        +-------------------+-------------------+------------------+
        | cropped_left      | cropped_left      |  first_event     |
        +-------------------+-------------------+------------------+
        | cropped_right     | cropped_right     |  last_event      |
        +-------------------+-------------------+------------------+

    See Also
    --------
    .TimedeltaHist : Plot the distribution of the time deltas between two events.
    .UserLifetimeHist : Plot the distribution of user lifetimes.

    Raises
    ------
    ValueError
        If both of ``left_cutoff`` and ``right_cutoff`` are empty.

    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.
    """

    params: LabelCroppedPathsParams

    @time_performance(
        scope="label_cropped_paths",
        event_name="init",
    )
    def __init__(self, params: LabelCroppedPathsParams):
        super().__init__(params=params)

    @time_performance(  # type: ignore
        scope="label_cropped_paths",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        user_col = schema.user_id
        time_col = schema.event_timestamp
        type_col = schema.event_type
        event_col = schema.event_name

        left_cutoff, left_truncated_unit = None, None
        right_cutoff, right_truncated_unit = None, None

        if self.params.left_cutoff:
            left_cutoff, left_truncated_unit = self.params.left_cutoff

        if self.params.right_cutoff:
            right_cutoff, right_truncated_unit = self.params.right_cutoff

        labeled_events = pd.DataFrame()

        if not left_cutoff and not right_cutoff:
            raise ValueError("Either left_cutoff or right_cutoff must be specified!")

        userpath = df.groupby(user_col)[time_col].agg(start=np.min, end=np.max)  # type: ignore

        if left_cutoff:
            timedelta = (userpath["end"] - df[time_col].min()) / np.timedelta64(1, left_truncated_unit)  # type: ignore
            cropped_users_index = userpath[timedelta < left_cutoff].index
            left_labeled_events = df.groupby(user_col).first().reindex(cropped_users_index).reset_index()

            left_labeled_events[event_col] = "cropped_left"
            left_labeled_events[type_col] = "cropped_left"
            labeled_events = pd.concat([labeled_events, left_labeled_events])

        if right_cutoff:
            timedelta = (df[time_col].max() - userpath["start"]) / np.timedelta64(
                1, right_truncated_unit  # type: ignore
            )
            cropped_users_index = userpath[timedelta < right_cutoff].index
            right_labeled_events = df.groupby(user_col).last().reindex(cropped_users_index).reset_index()

            right_labeled_events[event_col] = "cropped_right"
            right_labeled_events[type_col] = "cropped_right"
            labeled_events = pd.concat([labeled_events, right_labeled_events])

        result = pd.concat([df, labeled_events])

        collect_data_performance(
            scope="label_cropped_paths",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            not_hash_values=["left_cutoff", "right_cutoff"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



from .utils import RETE_CONFIG
from .__version__ import __version__
from .utils import check_version, supress_warnings
from . import (
    datasets,
    eventstream,
    preprocessing_graph,
    params_model,
    preprocessor,
    tooling,
    widget,
    data_processor,
    data_processors_lib,
)

__all__ = (
    "datasets",
    "eventstream",
    "preprocessing_graph",
    "params_model",
    "preprocessor",
    "tooling",
    "widget",
    "data_processor",
    "data_processors_lib",
    "__version__",
    "RETE_CONFIG",
)



from __future__ import annotations

from typing import Callable, Dict, List, Optional

from pandas import DataFrame, Series
from typing_extensions import (  # required for pydantic and python < 3.9.2
    NotRequired,
    Required,
    TypedDict,
)

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamSchemaType, EventstreamType

EventstreamFilter = Callable[[DataFrame, Optional[EventstreamSchemaType]], DataFrame]
GroupingRulesDict = Dict[str, EventstreamFilter]


class GroupEventsRule(TypedDict, total=False):
    event_name: Required[str]
    func: Required[EventstreamFilter]
    event_type: NotRequired[str]


class GroupEventsBulkHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(  # type: ignore
        scope="group_events_bulk",
        event_name="helper",
        event_value="combine",
    )
    def group_events_bulk(
        self: EventstreamType,
        grouping_rules: List[GroupEventsRule] | GroupingRulesDict,
        ignore_intersections: bool = False,
    ) -> EventstreamType:
        """
        Apply multiple grouping rules simultaneously.
        See also :py:meth:`GroupEvents<retentioneering.data_processors_lib.group_events.GroupEvents>`

        Parameters
        ----------
            %(GroupEventsBulk.parameters)s

        Returns
        -------
        Eventstream
            ``Eventstream`` with the grouped events according to the given grouping rules.

        """
        # avoid circular import
        from retentioneering.data_processors_lib import (
            GroupEventsBulk,
            GroupEventsBulkParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        params = GroupEventsBulkParams(grouping_rules=grouping_rules, ignore_intersections=ignore_intersections)  # type: ignore

        calling_params = params.dict()

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(processor=GroupEventsBulk(params=params))
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="map_events",
            event_name="metadata",
            called_params=calling_params,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



from __future__ import annotations

from typing import Dict, TypedDict, Union

from .column import ColumnId


class ThresholdValue(TypedDict):
    min: float | int
    max: float | int


ThresholdValueMap = Dict[ColumnId, ThresholdValue]

Threshold = Dict[ColumnId, Union[float, int]]

ThresholdWithFallback = Dict[str, Union[ThresholdValue, float, int]]



from __future__ import annotations

from typing import List, Literal, Union

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class LabelNewUsersHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(
        scope="label_new_users",
        event_name="helper",
        event_value="combine",
    )
    def label_new_users(self: EventstreamType, new_users_list: Union[List[int], Literal["all"]]) -> EventstreamType:
        """
        A method of ``Eventstream`` class that creates one
        of the synthetic events in each user's path: ``new_user`` or ``existing_user`` .

        Parameters
        ----------
            %(LabelNewUsers.parameters)s

        Returns
        -------
        Eventstream
             Input ``eventstream`` with new synthetic events.


        """
        calling_params = {"new_users_list": new_users_list}

        # avoid circular import
        from retentioneering.data_processors_lib import (
            LabelNewUsers,
            LabelNewUsersParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        node = EventsNode(
            processor=LabelNewUsers(params=LabelNewUsersParams(new_users_list=new_users_list))  # type: ignore
        )
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="label_new_users",
            event_name="metadata",
            called_params=calling_params,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



from typing import Optional

from .base import BaseReteException


class WidgetParseError(BaseReteException):
    field_name: Optional[str]

    def __init__(self, *args: object) -> None:
        self.field_name = None
        super().__init__(*args)


class ParseReteFuncError(WidgetParseError):
    pass



from __future__ import annotations

from hashlib import sha256

import pandas as pd


def hash_value(value: str) -> str:
    return sha256(value.encode("utf-8")).hexdigest()


def hash_dataframe(data: pd.DataFrame | pd.Series) -> str:
    return sha256(
        pd.util.hash_pandas_object(pd.concat([data.iloc[:1000], data.iloc[-1000:]]), index=True).values  # type: ignore
    ).hexdigest()



from __future__ import annotations

from os.path import dirname

import pandas as pd

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.eventstream import Eventstream, EventstreamSchema, RawDataSchema

module_path = dirname(__file__)


@time_performance(  # type: ignore
    scope="dataset",
    event_name="load_simple_shop",
    event_value="load_simple_shop",
)
def load_simple_shop(as_dataframe: bool = False, add_start_end_events: bool = True) -> pd.DataFrame | Eventstream:
    """
    Load a `simple_shop` demonstration dataset.

    Parameters
    ----------
    as_dataframe : bool, default False
        - If ``False`` the dataset is returned as eventstream
        - If ``True`` the dataset is returned as pandas.DataFrame
    """
    params = {"as_dataframe": as_dataframe, "add_start_end_events": add_start_end_events}

    df = pd.read_csv(module_path + "/data/simple-onlineshop.csv")
    if as_dataframe:
        collect_data_performance(
            scope="dataset",
            event_name="metadata",
            called_params=params,
            performance_data=None,
            eventstream_index=None,
        )

        return df
    else:
        stream = Eventstream(
            raw_data=df,
            raw_data_schema=RawDataSchema(event_name="event", event_timestamp="timestamp", user_id="user_id"),
            schema=EventstreamSchema(),
            add_start_end_events=add_start_end_events,
        )
        collect_data_performance(
            scope="dataset",
            event_name="metadata",
            called_params=params,
            performance_data=None,
            eventstream_index=stream._eventstream_index,
        )
        return stream


@time_performance(
    scope="dataset",
    event_name="load_simple_ab_test",
    event_value="load_simple_ab_test",
)
def load_simple_ab_test() -> pd.DataFrame:
    data = pd.read_csv(module_path + "/data/ab_test_demo.csv")
    data["transaction_ID"] = data["transaction_ID"].astype(str)
    data.loc[data["transaction_ID"] == "nan", "transaction_ID"] = None  # type: ignore
    data.loc[data["transaction_ID"].notna(), "transaction_ID"] = data.loc[
        data["transaction_ID"].notna(), "transaction_ID"
    ].apply(lambda x: x.replace(".0", ""))
    return data



from .add_negative_events import AddNegativeEvents, AddNegativeEventsParams
from .add_positive_events import AddPositiveEvents, AddPositiveEventsParams
from .add_segment import AddSegment, AddSegmentParams
from .add_start_end_events import AddStartEndEvents, AddStartEndEventsParams
from .collapse_loops import CollapseLoops, CollapseLoopsParams
from .drop_paths import DropPaths, DropPathsParams
from .drop_segment import DropSegment, DropSegmentParams
from .filter_events import FilterEvents, FilterEventsParams
from .group_events import GroupEvents, GroupEventsParams
from .group_events_bulk import GroupEventsBulk, GroupEventsBulkParams, GroupEventsRule
from .label_cropped_paths import LabelCroppedPaths, LabelCroppedPathsParams
from .label_lost_users import LabelLostUsers, LabelLostUsersParams
from .label_new_users import LabelNewUsers, LabelNewUsersParams
from .materialize_segment import MaterializeSegment, MaterializeSegmentParams
from .pipe import Pipe, PipeParams
from .remap_segment import RemapSegment, RemapSegmentParams
from .rename import RenameParams, RenameProcessor
from .rename_segment import RenameSegment, RenameSegmentParams
from .split_sessions import SplitSessions, SplitSessionsParams
from .truncate_paths import TruncatePaths, TruncatePathsParams



from __future__ import annotations

from typing import Optional, Tuple

from retentioneering.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from retentioneering.common.constants import DATETIME_UNITS
from retentioneering.utils.doc_substitution import docstrings

from ..types import EventstreamType


class LabelCroppedPathsHelperMixin:
    @docstrings.with_indent(12)
    @time_performance(
        scope="label_cropped_paths",
        event_name="helper",
        event_value="combine",
    )
    def label_cropped_paths(
        self: EventstreamType,
        left_cutoff: Optional[Tuple[float, DATETIME_UNITS]],
        right_cutoff: Optional[Tuple[float, DATETIME_UNITS]],
    ) -> EventstreamType:
        """
        A method of ``Eventstream`` class that creates new synthetic event(s) for each user based
        on the timeout threshold: ``cropped_left`` and ``cropped_right``.


        Parameters
        ----------
            %(LabelCroppedPaths.parameters)s

        Returns
        -------
        Eventstream
            Input ``eventstream`` with new synthetic events.


        """
        calling_params = {
            "left_cutoff": left_cutoff,
            "right_cutoff": right_cutoff,
        }
        not_hash_values = ["left_cutoff", "right_cutoff"]

        # avoid circular import
        from retentioneering.data_processors_lib import (
            LabelCroppedPaths,
            LabelCroppedPathsParams,
        )
        from retentioneering.preprocessing_graph import PreprocessingGraph
        from retentioneering.preprocessing_graph.nodes import EventsNode

        p = PreprocessingGraph(source_stream=self)  # type: ignore

        params = dict(left_cutoff=left_cutoff, right_cutoff=right_cutoff)

        node = EventsNode(processor=LabelCroppedPaths(params=LabelCroppedPathsParams(**params)))  # type: ignore
        p.add_node(node=node, parents=[p.root])
        result = p.combine(node)
        del p
        collect_data_performance(
            scope="label_cropped_paths",
            event_name="metadata",
            called_params=calling_params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=result._eventstream_index,
        )

        return result



import warnings

import urllib3
from numba import NumbaDeprecationWarning

warnings.simplefilter(action="ignore", category=NumbaDeprecationWarning)
urllib3.disable_warnings()

import logging

logging.getLogger("requests").setLevel(logging.NOTSET)
logging.getLogger("urllib3").setLevel(logging.NOTSET)



from __future__ import annotations

from typing import Any, Callable, List

import pandas as pd

from retentioneering.backend.tracker import collect_data_performance, time_performance
from retentioneering.data_processor import DataProcessor
from retentioneering.eventstream.schema import EventstreamSchema
from retentioneering.eventstream.types import EventstreamSchemaType, EventstreamType
from retentioneering.params_model import ParamsModel
from retentioneering.utils.doc_substitution import docstrings
from retentioneering.utils.hash_object import hash_dataframe
from retentioneering.widget.widgets import ListOfString, ReteFunction

EventstreamFilter = Callable[[pd.DataFrame, EventstreamSchema], Any]


def _default_func(eventstream: EventstreamType, targets: List[str]) -> pd.DataFrame:
    """
    Filter rows with target events from the input eventstream.

    Parameters
    ----------
    eventstream : Eventstream
        Source eventstream or output from previous nodes.

    targets : list of str
        Each event from that list is associated with the bad result (scenario)
        of user's behaviour (experience) in the product.
        If there are several target events in user path - the event with minimum timestamp is taken.

    Returns
    -------
    pd.DataFrame
        Filtered DataFrame with targets and its timestamps.
    """
    user_col = eventstream.schema.user_id
    time_col = eventstream.schema.event_timestamp
    event_col = eventstream.schema.event_name
    df = eventstream.to_dataframe()

    targets_index = df[df[event_col].isin(targets)].groupby(user_col)[time_col].idxmin()  # type: ignore

    return df.loc[targets_index]  # type: ignore


class AddNegativeEventsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.AddNegativeEvents` class.
    """

    targets: List[str]
    # @TODO: remove eventstream from the "func" signature in a future major release. Aleksei Avramenko
    func: Callable = _default_func

    _widgets = {"func": ReteFunction(), "targets": ListOfString()}


@docstrings.get_sections(base="AddNegativeEvents")  # type: ignore
class AddNegativeEvents(DataProcessor):
    """
    Create new synthetic events in paths of all users having the specified event(s):
    ``negative_target_RAW_EVENT_NAME``.

    Parameters
    ----------
    targets : list of str
        Define the list of events that we consider negative.
        If there are several target events in the user path, the event with the minimum timestamp is taken.
    func : Callable, default _default_func_negative
        Filter rows with target events from the input eventstream.

    Returns
    -------
    Eventstream
        ``Eventstream`` with new synthetic events only added to the users who fit the conditions.

        +--------------------------------+-----------------+-----------------------------+
        | **event_name**                 | **event_type**  | **timestamp**               |
        +--------------------------------+-----------------+-----------------------------+
        | negative_target_RAW_EVENT_NAME | negative_target | min(targets)                |
        +--------------------------------+-----------------+-----------------------------+

    Notes
    -----
    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.


    """

    params: AddNegativeEventsParams

    @time_performance(
        scope="add_negative_events",
        event_name="init",
    )
    def __init__(self, params: AddNegativeEventsParams):
        super().__init__(params=params)

    @time_performance(
        scope="add_negative_events",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        from retentioneering.eventstream.eventstream import Eventstream

        type_col = schema.event_type
        event_col = schema.event_name

        func = self.params.func
        targets = self.params.targets

        eventstream = Eventstream(
            raw_data_schema=schema.to_raw_data_schema(event_index=True),
            raw_data=df,
            add_start_end_events=False,
        )

        # @TODO: remove eventstream from the "func" signature in a future major release. Aleksei Avramenko
        negative_targets: pd.DataFrame = func(eventstream, targets)
        negative_targets[type_col] = "negative_target"
        negative_targets[event_col] = "negative_target_" + negative_targets[event_col]

        result = pd.concat([eventstream.to_dataframe(), negative_targets])

        collect_data_performance(
            scope="add_negative_events",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            performance_data={
                "parent": {
                    "shape": df.shape,
                    "hash": hash_dataframe(df),
                },
                "child": {
                    "shape": result.shape,
                    "hash": hash_dataframe(result),
                },
            },
        )

        return result



from __future__ import annotations

from abc import ABCMeta, abstractmethod

from retentioneering.backend.tracker.tracking_info import TrackingInfo


class ConnectorProtocol(metaclass=ABCMeta):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def send_message(self, data: TrackingInfo) -> None: ...



from __future__ import annotations

from typing import Any, Dict, List, Union

import pandas as pd

from retentioneering.tooling.transition_graph.types import (
    ThresholdValue,
    ThresholdValueMap,
)

RenameRule = Dict[str, Union[List[str], str]]

SerializedNodelist = List[Dict[str, Any]]

IS_DISABLED_COL = "is_disabled"
IS_PINNED_COL = "is_pinned"
PARENT_ID_COL = "parent_id"
IS_GROUP_COL = "is_group"
IS_AGGREGATED_COL = "is_aggregated"
NAME_COL = "name"
IS_OUT_OF_THRESHOLD_COL = "rete_is_out_of_threshold"


class Nodelist:
    nodelist_df: pd.DataFrame
    event_col: str
    time_col: str
    weight_cols: list[str]
    payload_df: pd.DataFrame | None

    def __init__(
        self, event_col: str, time_col: str, weight_cols: list[str] | None, nodelist_df: pd.DataFrame | None = None
    ) -> None:
        self.event_col = event_col
        self.time_col = time_col
        self.weight_cols = weight_cols if weight_cols is not None else []

        if nodelist_df is not None:
            self.validate_nodelist(nodelist_df)
            self.nodelist_df = nodelist_df
        else:
            self.nodelist_df = self.create_empty_nodelist_df()

    def get_threshold_min_max(self) -> ThresholdValueMap:
        result: ThresholdValueMap = {}

        for weight_col in self.weight_cols:
            min_value: float | int = self.nodelist_df[weight_col].min()
            max_value: float | int = self.nodelist_df[weight_col].max()
            result[weight_col] = {
                "min": min_value,
                "max": max_value,
            }

        return result

    def fit_threshold(self, threshold: ThresholdValueMap, prev_min_max: ThresholdValueMap) -> ThresholdValueMap:
        new_threshold = threshold.copy()
        threshold_min_max = self.get_threshold_min_max()

        for weight_col, value in new_threshold.items():
            if weight_col not in prev_min_max:
                continue
            if weight_col not in threshold_min_max:
                continue

            if value["max"] == prev_min_max[weight_col]["max"]:
                new_threshold[weight_col]["max"] = threshold_min_max[weight_col]["max"]

        return new_threshold

    def calculate_nodelist(
        self,
        data: pd.DataFrame,
        nodes_thresholds: ThresholdValueMap | None = None,
    ) -> pd.DataFrame:
        res: pd.DataFrame = data.groupby([self.event_col])[self.time_col].count().reset_index()

        for weight_col in self.weight_cols:
            by_col = data.groupby([self.event_col])[weight_col].nunique().reset_index()
            res = res.join(by_col[weight_col])
            res[weight_col] = res[weight_col].astype(float)

        res = res.sort_values(by=self.time_col, ascending=False)
        res = res.drop(columns=[self.time_col], axis=1)

        # just update weights
        res.set_index(self.event_col, inplace=True)
        self.nodelist_df.set_index(self.event_col, inplace=True)

        self.nodelist_df.update(res)
        res.reset_index(inplace=True)
        self.nodelist_df.reset_index(inplace=True)

        new_nodes = res[~res[self.event_col].isin(self.nodelist_df[self.event_col])]
        new_nodes = self.set_default_columns(new_nodes)

        if self.nodelist_df.empty:
            self.nodelist_df = new_nodes
        elif not new_nodes.empty:
            self.nodelist_df = pd.concat([self.nodelist_df, new_nodes])

        self.update_threshold(nodes_thresholds=nodes_thresholds)

        return self.nodelist_df

    def set_default_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        df[IS_DISABLED_COL] = False
        df[IS_PINNED_COL] = False
        df[PARENT_ID_COL] = None
        df[IS_GROUP_COL] = False
        df[IS_AGGREGATED_COL] = False
        df[NAME_COL] = df[self.event_col]
        df[IS_OUT_OF_THRESHOLD_COL] = False

        df[IS_DISABLED_COL] = df[IS_DISABLED_COL].astype(bool)
        df[IS_PINNED_COL] = df[IS_PINNED_COL].astype(bool)
        df[IS_GROUP_COL] = df[IS_GROUP_COL].astype(bool)
        df[IS_AGGREGATED_COL] = df[IS_AGGREGATED_COL].astype(bool)
        df[IS_OUT_OF_THRESHOLD_COL] = df[IS_OUT_OF_THRESHOLD_COL].astype(bool)

        return df

    def create_empty_nodelist_df(self) -> pd.DataFrame:
        columns = (
            [self.event_col]
            + self.weight_cols
            + [
                IS_DISABLED_COL,
                IS_PINNED_COL,
                PARENT_ID_COL,
                IS_GROUP_COL,
                IS_AGGREGATED_COL,
                NAME_COL,
                IS_OUT_OF_THRESHOLD_COL,
            ]
        )

        df = pd.DataFrame(columns=columns)

        df[IS_DISABLED_COL] = df[IS_DISABLED_COL].astype(bool)
        df[IS_PINNED_COL] = df[IS_PINNED_COL].astype(bool)
        df[IS_GROUP_COL] = df[IS_GROUP_COL].astype(bool)
        df[IS_AGGREGATED_COL] = df[IS_AGGREGATED_COL].astype(bool)
        df[IS_OUT_OF_THRESHOLD_COL] = df[IS_OUT_OF_THRESHOLD_COL].astype(bool)

        return df

    def is_valid_nodelist_df(self, df: pd.DataFrame) -> bool:
        required_columns = [self.event_col] + [
            IS_DISABLED_COL,
            IS_PINNED_COL,
            PARENT_ID_COL,
            IS_GROUP_COL,
            IS_AGGREGATED_COL,
            NAME_COL,
        ]
        return all(column in df.columns for column in required_columns)

    def validate_nodelist(self, df: pd.DataFrame) -> None:
        if not self.is_valid_nodelist_df(df):
            raise ValueError("invalid nodelist_df")

    def groups_to_rename_rules(self) -> list[RenameRule]:
        rename_rules: list[RenameRule] = []
        df = self.nodelist_df
        groups: list[str] = df.loc[
            (df[IS_GROUP_COL] == True) & (df[IS_AGGREGATED_COL] == True), self.event_col
        ].tolist()

        for group in groups:
            child_events = df.loc[df[PARENT_ID_COL] == group, self.event_col].tolist()
            if len(child_events) > 0:
                rename_rules.append(dict(group_name=group, child_events=child_events))

        return rename_rules

    def renamed_events_to_rename_rules(self) -> list[RenameRule]:
        rename_rules: list[RenameRule] = []

        df = self.get_ungrouped_nodes()
        renamed = df[df[self.event_col] != df[NAME_COL]]

        for _, row in renamed.iterrows():
            source_name = row[self.event_col]
            alias = row[NAME_COL]
            rename_rules.append(dict(group_name=alias, child_events=[source_name]))

        return rename_rules

    def update(self, df: pd.DataFrame) -> None:
        self.payload_df = df
        self.validate_nodelist(df)
        df[IS_OUT_OF_THRESHOLD_COL] = False
        cols = [self.event_col] + [
            IS_DISABLED_COL,
            IS_PINNED_COL,
            PARENT_ID_COL,
            IS_GROUP_COL,
            IS_AGGREGATED_COL,
            NAME_COL,
            IS_OUT_OF_THRESHOLD_COL,
        ]
        update_set = df[cols]

        groups_mask = update_set[IS_GROUP_COL] == True
        groups = update_set[groups_mask]

        self.nodelist_df.set_index(self.event_col, inplace=True)
        self.nodelist_df.update(groups.set_index(self.event_col))
        self.nodelist_df.reset_index(inplace=True)

        new_groups_mask = ~groups[self.event_col].isin(self.nodelist_df[self.event_col])
        new_groups = groups[new_groups_mask]

        for weight_col in self.weight_cols:
            new_groups[weight_col] = 0

        if not new_groups.empty:
            self.nodelist_df = pd.concat([self.nodelist_df, new_groups])

        events_nodes_mask = update_set[IS_GROUP_COL] == False
        events_nodes = update_set[events_nodes_mask]
        self.nodelist_df.set_index(self.event_col, inplace=True)
        self.nodelist_df.update(events_nodes.set_index(self.event_col))
        self.nodelist_df.reset_index(inplace=True)

    def update_threshold(self, nodes_thresholds: ThresholdValueMap | None = None) -> None:
        # filter nodes by threshold
        if nodes_thresholds:

            def is_out_of_threshold(row: pd.Series, column_name: str, threshold_value: ThresholdValue) -> bool:
                value = row[column_name]
                return not threshold_value["min"] <= value <= threshold_value["max"]

            self.nodelist_df[IS_OUT_OF_THRESHOLD_COL] = False

            for threshold_col, threshold in nodes_thresholds.items():
                if threshold_col in self.nodelist_df.columns:
                    self.nodelist_df[IS_OUT_OF_THRESHOLD_COL] = (
                        self.nodelist_df.apply(is_out_of_threshold, axis=1, args=(threshold_col, threshold))
                        | self.nodelist_df[IS_OUT_OF_THRESHOLD_COL]
                    )

    def get_ungrouped_nodes(self) -> pd.DataFrame:
        df = self.nodelist_df

        groups_roots_mask = (df[IS_GROUP_COL] == True) & (df[IS_AGGREGATED_COL] == True)
        not_grouped_mask = (df[IS_GROUP_COL] == False) & (df[PARENT_ID_COL].isna())

        parents = df.set_index(self.event_col)

        def parent_condition(
            row: pd.Series,
        ) -> bool:
            if pd.notna(row[PARENT_ID_COL]):
                parent = parents.loc[row[PARENT_ID_COL]]
                return parent[IS_GROUP_COL] == True and parent[IS_AGGREGATED_COL] == False
            return False

        grouped_but_not_aggregated_mask = (
            (df[IS_GROUP_COL] == False) & (df[PARENT_ID_COL].notna()) & df.apply(parent_condition, axis=1)
        )

        mask = groups_roots_mask | not_grouped_mask | grouped_but_not_aggregated_mask
        return df[mask]

    def get_out_of_threshold_nodes(self, only_ungrouped: bool = False) -> list[str]:
        df = self.get_ungrouped_nodes() if only_ungrouped else self.nodelist_df
        return df.loc[(df[IS_OUT_OF_THRESHOLD_COL] == True) & (df[IS_PINNED_COL] != True), self.event_col].tolist()

    def get_disabled_nodes(self, groups: bool = False) -> list[str]:
        df = self.nodelist_df
        if not groups:
            df = df[df[IS_GROUP_COL] == False]

        return df.loc[(df[IS_DISABLED_COL] == True) & (df[IS_PINNED_COL] != True), self.event_col].to_list()

    def get_min_max(self) -> ThresholdValueMap:
        result: ThresholdValueMap = {}

        if self.nodelist_df is None:  # type: ignore
            return result

        for weight_col in self.weight_cols:
            min_value = float(self.nodelist_df[weight_col].min())
            max_value = float(self.nodelist_df[weight_col].max())

            result[weight_col] = {"min": min_value, "max": max_value}

        return result

    def to_dict(self) -> SerializedNodelist:
        return self.nodelist_df.to_dict(orient="records")  # type: ignore

    def copy(self) -> Nodelist:
        weight_cols = self.weight_cols.copy()
        nodelist_df = self.nodelist_df.copy()

        return Nodelist(
            event_col=self.event_col,
            time_col=self.time_col,
            weight_cols=weight_cols,
            nodelist_df=nodelist_df,
        )
